import scarletio

from scarletio import *
from scarletio.http_client import *
from scarletio.websocket import *
from scarletio.web_common import *


from scarletio import http_client
from scarletio import websocket
from scarletio import web_common

__all__ = (
    *scarletio.__all__,
    *http_client.__all__,
    *websocket.__all__,
    *web_common.__all__,
)

import warnings

warnings.warn(
    (
        f'Hata\'s backend has been moved out to a separate library.\n'
        f'Please import from `scarletio` instead.'
    ),
    FutureWarning,
    stacklevel = 3,
)



__all__ = ('ChannelMetadataGuildVoice',)

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import (
    PERMISSION_MASK_CONNECT, PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_STAGE_DENY,
    PERMISSION_VOICE_DENY_CONNECTION
)
from ...preconverters import preconvert_bool, preconvert_preinstanced_type

from .. import channel_types as CHANNEL_TYPES
from ..preinstanced import VideoQualityMode

from .guild_main_base import ChannelMetadataGuildMainBase
from .guild_voice_base import ChannelMetadataGuildVoiceBase


class ChannelMetadataGuildVoice(ChannelMetadataGuildVoiceBase):
    """
    Guild voice channel metadata.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    bitrate : `int`
        The bitrate (in bits) of the voice channel.
    region : `None`, ``VoiceRegion``
        The voice region of the channel.
    user_limit : `int`
        The maximal amount of users, who can join the voice channel, or `0` if unlimited.
    nsfw : `bool`
        Whether the channel is marked as non safe for work.
    video_quality_mode : ``VideoQualityMode``
        The video quality of the voice channel.
    
    Class Attributes
    ----------------
    type : `int` = `CHANNEL_TYPES.guild_voice`
        The channel's type.
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('nsfw', 'video_quality_mode',)
    
    type = CHANNEL_TYPES.guild_voice
    
    @copy_docs(ChannelMetadataGuildVoiceBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if not ChannelMetadataGuildVoiceBase._compare_attributes_to(self, other):
            return False
        
        if self.nsfw != other.nsfw:
            return False
        
        if self.video_quality_mode is not other.video_quality_mode:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildVoice, cls)._create_empty()
        
        self.nsfw = False
        self.video_quality_mode = VideoQualityMode.none
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildVoiceBase._update_attributes(self, data)
        
        self.nsfw = data.get('nsfw', False)
        
        self.video_quality_mode = VideoQualityMode.get(data.get('video_quality_mode', 1))
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildVoiceBase._difference_update_attributes(self, data)
        
        nsfw = data.get('nsfw', False)
        if self.nsfw != nsfw:
            old_attributes['nsfw'] = self.nsfw
            self.nsfw = nsfw
        
        video_quality_mode = VideoQualityMode.get(data.get('video_quality_mode', 1))
        if self.video_quality_mode is not video_quality_mode:
            old_attributes['video_quality_mode'] = self.video_quality_mode
            self.video_quality_mode = video_quality_mode
        
        return old_attributes
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase._precreate)
    def _precreate(cls, keyword_parameters):
        self = super(ChannelMetadataGuildVoice, cls)._precreate(keyword_parameters)
        
        try:
            nsfw = keyword_parameters.pop('nsfw')
        except KeyError:
            pass
        else:
            nsfw = preconvert_bool(nsfw, 'nsfw')
            self.nsfw = nsfw
        
        try:
            video_quality_mode = keyword_parameters.pop('video_quality_mode')
        except KeyError:
            pass
        else:
            video_quality_mode = preconvert_preinstanced_type(
                video_quality_mode,
                'video_quality_mode',
                VideoQualityMode,
            )
            
            self.video_quality_mode = video_quality_mode
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._to_data)
    def _to_data(self):
        data = ChannelMetadataGuildVoiceBase._to_data(self)
        
        # nsfw
        if self.nsfw:
            data['nsfw'] = True
        
        video_quality_mode = self.video_quality_mode
        if (video_quality_mode is not VideoQualityMode.nonde):
            data['video_quality_mode'] = video_quality_mode.value
        
        return data


    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have text permissions
        result &= PERMISSION_STAGE_DENY
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have text permissions
        result &= PERMISSION_STAGE_DENY
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)



__all__ = ('ContextCommand',)

from scarletio import copy_docs

from .....discord.events.handling_helpers import Router, check_name, route_name, route_value

from ...converters import get_context_command_parameter_converters
from ...utils import _check_maybe_route, raw_name_to_display
from ...wrappers import CommandWrapper
from ...exceptions import handle_command_exception
from ...responding import process_command_coroutine

from ..command_base_application_command import CommandBaseApplicationCommand
from ..command_base_application_command.helpers import (
    _validate_allow_by_default, _validate_allow_in_dm, _validate_delete_on_unload, _validate_guild,
    _validate_is_global, _validate_name, _validate_required_permissions
)
from ..helpers import validate_application_target_type
from ...response_modifier import ResponseModifier


class ContextCommand(CommandBaseApplicationCommand):
    """
    Base class for ``Slasher``'s application commands.
    
    Attributes
    ----------
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parent_reference : `None`, ``WeakReferer`` to ``Slasher``
        Reference to the slasher application command's parent.
    
    name : `str`
        Application command name. It's length can be in range [1:32].
    
    _permission_overwrites : `None`, `dict` of (`int`, `list` of ``ApplicationCommandPermissionOverwrite``)
        Permission overwrites applied to the context command.

    _registered_application_command_ids : `None`, `dict` of (`int`, `int`) items
        The registered application command ids, which are matched by the command's schema.
        
        If empty set as `None`, if not then the keys are the respective guild's id and the values are the application
        command id.
    
    _schema : `None`, ``ApplicationCommand``
        Internal slot used by the ``.get_schema`` method.
    
    _unloading_behaviour : `int`
        Behaviour what describes what should happen when the command is removed from the slasher.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_DELETE    | 0     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_KEEP      | 1     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_INHERIT   | 2     |
        +-------------------------------+-------+
    
    allow_by_default : `None`, `bool`
        Whether the command is enabled by default for everyone who has `use_application_commands` permission.

    guild_ids : `None`, `set` of `int`
        The ``Guild``'s id to which the command is bound to.
    
    is_global : `bool`
        Whether the command is a global command.
        
        Guild commands have ``.guild_ids`` set as `None`.
    
    required_permissions : `None`, ``Permission``
        The required permissions to use the application command inside of a guild.
    
    _command_function : `async-callable˛
        The command's function to call.
    
    _parameter_converters : `tuple` of ``ParameterConverter``
        Parsers to parse command parameters.
    
    response_modifier : `None`, ``ResponseModifier``
        Modifies values returned and yielded to command coroutine processor.
    
    target : ``ApplicationCommandTargetType``
        The target type of the context command.
    
    Class Attributes
    ----------------
    COMMAND_COMMAND_NAME : `str`
        The command's name defining parameter's name.
    COMMAND_PARAMETER_NAMES : `tuple` of `str`
        All parameters names accepted by ``.__new__``
    COMMAND_NAME_NAME : `str`
        The command's "command" defining parameter's name.
    
    description : `None` = `None`
        The command's description.
    """
    __slots__ = ('_command_function', '_parameter_converters', 'response_modifier', 'target',)
    
    COMMAND_PARAMETER_NAMES = (
        *CommandBaseApplicationCommand.COMMAND_PARAMETER_NAMES,
        'target',
    )
    
    
    def __new__(
        cls, func, name=None, is_global=None, guild=None, is_default=None, delete_on_unload=None,
        allow_by_default=None, allow_in_dm=None, required_permissions=None, target=None, **keyword_parameters,
    ):
        """
        Creates a new ``SlashCommand`` with the given parameters.
        
        Parameters
        ----------
        func : `None`, `async-callable` = `None`, Optional
            The function used as the command when using the respective context command.
        
        name : `None`, `str`, `tuple` of (`str`, `Ellipsis`, `None`) = `None`, Optional
            The command's name if applicable. If not given or if given as `None`, the `func`'s name will be use
            instead.
        
        is_global : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the context command is global. Defaults to `False`.
        
        guild : `None`, ``Guild``,  `int`, (`list`, `set`) of (`int`, ``Guild``) or \
                `tuple` of (`None`, ``Guild``,  `int`, `Ellipsis`, (`list`, `set`) of (`int`, ``Guild``)) = `None`
                , Optional
            To which guild(s) the command is bound to.
        
        is_global : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the context command is the default command in it's category.
        
        delete_on_unload : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the command should be deleted from Discord when removed.
        
        allow_by_default : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the command is enabled by default for everyone who has `use_application_commands` permission.
        
        allow_in_dm : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the command can be used in private channels (dm).
        
        required_permissions : `None`, `int`, ``Permission``, `tuple` of (`None`, `int`, ``Permission``,
                `Ellipsis`) = `None`, Optional
            The required permissions to use the application command inside of a guild.
        
        target : `None`, `int`, `str`, ``ApplicationCommandTargetType`` = `None`, Optional
            The target type of the command.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other parameters
        ----------------
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, ``AllowedMentionProxy``, \
                `list` of (`str`, ``UserBase``, ``Role`` ), Optional (Keyword only)
            Which user or role can the response message ping (or everyone).
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the response message should only be shown for the invoking user.
        wait_for_acknowledgement : `bool`, Optional (Keyword only)
            Whether acknowledge tasks should be ensure asynchronously.
        
        Returns
        -------
        self : ``SlashCommand``, ``Router``
        
        Raises
        ------
        TypeError
            If a parameter's type is incorrect.
        ValueError
            If a parameter's value is incorrect.
        """
        if (func is not None) and isinstance(func, CommandWrapper):
            command, wrappers = func.fetch_function_and_wrappers_back()
        else:
            command = func
            wrappers = None
        
        if command is None:
            raise ValueError(
                f'For context commands `command` parameter is required (cannot be `None` either).'
            )
        
        target = validate_application_target_type(target)
        
        # Check for routing
        
        route_to = 0
        name, route_to = _check_maybe_route('name', name, route_to, _validate_name)
        is_global, route_to = _check_maybe_route('is_global', is_global, route_to, _validate_is_global)
        guild_ids, route_to = _check_maybe_route('guild', guild, route_to, _validate_guild)
        unloading_behaviour, route_to = _check_maybe_route(
            'delete_on_unload', delete_on_unload, route_to, _validate_delete_on_unload
        )
        allow_by_default, route_to = _check_maybe_route(
            'allow_by_default', allow_by_default, route_to, _validate_allow_by_default
        )
        allow_in_dm, route_to = _check_maybe_route('allow_in_dm', allow_in_dm, route_to, _validate_allow_in_dm)
        required_permissions, route_to = _check_maybe_route(
            'required_permissions', required_permissions, route_to, _validate_required_permissions
        )
        
        if route_to:
            name = route_name(name, route_to)
            name = [raw_name_to_display(check_name(command, sub_name)) for sub_name in name]
            
            is_global = route_value(is_global, route_to)
            guild_ids = route_value(guild_ids, route_to)
            unloading_behaviour = route_value(unloading_behaviour, route_to)
            allow_by_default = route_value(allow_by_default, route_to)
            allow_in_dm = route_value(allow_in_dm, route_to)
            required_permissions = route_value(required_permissions, route_to)
            target = route_value(target, route_to)
        
        else:
            name = check_name(command, name)
            name = raw_name_to_display(name)
            
        
        # Check extra parameters
        response_modifier = ResponseModifier(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        
        command, parameter_converters = get_context_command_parameter_converters(command)
        
        
        if route_to:
            router = []
            
            for (
                name, is_global, guild_ids, unloading_behaviour, allow_by_default,
                required_permissions, allow_in_dm
            ) in zip(
                name, is_global, guild_ids, unloading_behaviour, allow_by_default,
                required_permissions, allow_in_dm
            ):
                
                if is_global and (guild_ids is not None):
                    raise TypeError(
                        f'`is_global` and `guild` contradict each other, got is_global={is_global!r}, '
                        f'guild={guild!r}.'
                    )
                
                self = object.__new__(cls)
                self.guild_ids = guild_ids
                self.is_global = is_global
                self.name = name
                self._schema = None
                self._registered_application_command_ids = None
                self._unloading_behaviour = unloading_behaviour
                self.allow_by_default = allow_by_default
                self.allow_in_dm = allow_in_dm
                self.required_permissions = required_permissions
                self._permission_overwrites = None
                self.target = target
                self._exception_handlers = None
                self._parent_reference = None
                self._parameter_converters = parameter_converters
                self._command_function = command
                self.response_modifier = response_modifier
                
                if (wrappers is not None):
                    for wrapper in wrappers:
                        wrapper.apply(self)
                
                router.append(self)
            
            return Router(router)
        
        else:
            if is_global and (guild_ids is not None):
                raise TypeError(
                    f'`is_global` and `guild` contradict each other, got is_global={is_global!r}, '
                    f'guild={guild!r}.'
                )
            
            self = object.__new__(cls)
            self.guild_ids = guild_ids
            self.is_global = is_global
            self.name = name
            self._schema = None
            self._registered_application_command_ids = None
            self._unloading_behaviour = unloading_behaviour
            self.allow_by_default = allow_by_default
            self.allow_in_dm = allow_in_dm
            self.required_permissions = required_permissions
            self._permission_overwrites = None
            self.target = target
            self._exception_handlers = None
            self._parent_reference = None
            self._parameter_converters = parameter_converters
            self._command_function = command
            self.response_modifier = response_modifier
            
            if (wrappers is not None):
                for wrapper in wrappers:
                    wrapper.apply(self)
            
            return self
    
    
    @copy_docs(CommandBaseApplicationCommand._cursed_repr_builder)
    def _cursed_repr_builder(self):
        for repr_parts in CommandBaseApplicationCommand._cursed_repr_builder(self):
            
            yield repr_parts
            
            repr_parts.append(', target=')
            repr_parts.append(self.target.name)
    
    
    @copy_docs(CommandBaseApplicationCommand.invoke)
    async def invoke(self, client, interaction_event):
        parameters = []
        
        for parameter_converter in self._parameter_converters:
            try:
                parameter = await parameter_converter(client, interaction_event, None)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                exception = err
            
            else:
                parameters.append(parameter)
                continue
                
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
            return
            
        
        command_coroutine = self._command_function(*parameters)
        
        try:
            await process_command_coroutine(
                client,
                interaction_event,
                self.response_modifier,
                command_coroutine,
            )
        except GeneratorExit:
            raise
        
        except BaseException as err:
            exception = err
        
        else:
            return
        
        await handle_command_exception(
            self,
            client,
            interaction_event,
            exception,
        )
        return
    
    
    @copy_docs(CommandBaseApplicationCommand.copy)
    def copy(self):
        new = CommandBaseApplicationCommand.copy(self)
        
        # _command_function
        new._command_function = self._command_function
        
        # _parameter_converters
        new._parameter_converters = self._parameter_converters
        
        # response_modifier
        new.response_modifier = self.response_modifier
        
        # target
        new.target = self.target
        
        return new
    
    
    @copy_docs(CommandBaseApplicationCommand.__hash__)
    def __hash__(self):
        hash_value = CommandBaseApplicationCommand.__hash__(self)
        
        # _command_function
        command_function = self._command_function
        try:
            command_function_hash_value = hash(command_function)
        except TypeError:
            command_function_hash_value = object.__hash__(command_function)
        hash_value ^= command_function_hash_value
        
        # _parameter_converters
        # Internal field
        
        # response_modifier
        hash_value ^= hash(self.response_modifier)
        
        # target
        hash_value ^= self.target.value << 20
        
        return hash_value
    
    
    @copy_docs(CommandBaseApplicationCommand._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBaseApplicationCommand._is_equal_same_type(self, other):
            return False
        
        # _command_function
        if self._command_function != other._command_function:
            return False
        
        # _parameter_converters
        # Internal field
        
        # response_modifier
        if self.response_modifier != other.response_modifier:
            return False
        
        # target
        if self.target is not other.target:
            return False
        
        return True



__all__ = ('ComponentDynamic', )

import reprlib

from scarletio import RichAttributeErrorBaseType, copy_docs

from ...bases import PreinstancedBase
from ...emoji import create_partial_emoji_data, create_partial_emoji_from_data
from ...preconverters import preconvert_preinstanced_type

from .component_base import ComponentBase
from .preinstanced import ComponentType


def dynamic_component_style_serializer(style):
    if isinstance(style, PreinstancedBase):
        style = style.value
    
    return style


COMPONENT_DYNAMIC_SERIALIZERS = {
    'emoji': create_partial_emoji_data,
    'style': dynamic_component_style_serializer,
}

del dynamic_component_style_serializer

COMPONENT_DYNAMIC_DESERIALIZERS = {
    'emoji': create_partial_emoji_from_data,
}


COMPONENT_ATTRIBUTE_NAMES = frozenset((
    'components',
    'custom_id',
    'disabled',
    'emoji',
    'label',
    'max_length',
    'max_values',
    'min_length',
    'min_values',
    'options',
    'placeholder',
    'required',
    'style',
    'url',
    'value',
))


class ComponentDynamic(ComponentBase):
    """
    Dynamic component type for not implemented component models.
    
    Attributes
    ----------
    _data : `dict` of (`str`, `Any`)
        The dynamically stored attributes of the component.
    type : ``ComponentType``
        The component's type.
    """
    __slots__ = ('_data', 'type')
    
    def __new__(cls, type_, **kwargs):
        """
        Creates a new component instance.
        
        Parameters
        ----------
        type_ : ``ComponentType``, `int`
            The component's type.
        **kwargs : Keyword parameters
            Additional attributes of the component.
        """
        type_ = preconvert_preinstanced_type(type_, 'type_', ComponentType)
        
        self = object.__new__(cls)
        self.type = type_
        self._data = kwargs
        return self
    
    
    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.type = ComponentType.get(data['type'])
        
        validated_data = {}
        for key, value in data.items():
            if key == 'type':
                continue
            
            try:
                deserializer = COMPONENT_DYNAMIC_DESERIALIZERS[key]
            except KeyError:
                pass
            else:
                value = deserializer(value)
            
            validated_data[key] = value
        
        self._data = validated_data
        
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        data = {
            'type' : self.type.value
        }
        
        for key, value in self._data:
            try:
                serializer = COMPONENT_DYNAMIC_DESERIALIZERS[key]
            except KeyError:
                pass
            else:
                value = serializer(value)
            
            data[key] = value
        
        return data
    
    
    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__, ' type=']
        
        type_ = self.type
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        for key, value in self._data:
            if value is None:
                continue
            
            if isinstance(value, str):
                value = reprlib.repr(value)
            else:
                value = repr(value)
            
            repr_parts.append(', ')
            repr_parts.append(key)
            repr_parts.append('=')
            repr_parts.append(value)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.type = self.type
        new._data = self._data.copy()
        
        return new
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        > Dynamic component do not accepts any additional attributes, and returns just a copy of itself.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            Keyword parameters referencing attributes.
        
        Returns
        -------
        new : ``ComponentDynamic``
        """
        if kwargs:
            raise TypeError(f'Unused or unsettable attributes: {kwargs!r}.')
        
        return self.copy()
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        if self.type is not other.type:
            return False
        
        if self._data != other._data:
            return False
        
        return True
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        return object.__hash__(self)
    
    
    def __getattr__(self, attribute_name):
        """Returns the component's fields if applicable"""
        try:
            attribute_value = self._data[attribute_name]
        except KeyError:
            if attribute_name in COMPONENT_ATTRIBUTE_NAMES:
                attribute_value = None
            else:
                # Linter wont cry if we do a return or such.
                return RichAttributeErrorBaseType.__getattr__(self, attribute_name)
        
        return attribute_value
    
    
    def __dir__(self):
        """Returns the attributes of the component."""
        directory = set(object.__dir__(self))
        directory.update(self._data.keys())
        directory.update(COMPONENT_ATTRIBUTE_NAMES)
        return sorted(directory)



__all__ = ('ComponentBase',)

from scarletio import RichAttributeErrorBaseType, export

from .preinstanced import ComponentType


@export
class ComponentBase(RichAttributeErrorBaseType):
    """
    Base class for 3rd party components.
    
    Class Attributes
    ----------------
    custom_id : `NoneType` = `None`
        Placeholder for sub-classes without `custom_id` attribute.
    type : ``ComponentType`` = `ComponentType.none`
        The component's type.
    """
    __slots__ = ()
    
    custom_id = None
    type = ComponentType.none
    
    def __new__(cls):
        """
        Creates a component base instance.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new message component from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Message component data.
        
        Returns
        -------
        self : ``ComponentBase``
            The created component instance.
        """
        return None
    
    
    def to_data(self):
        """
        Converts the component to json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        # type
        data = {
            'type' : self.type.value
        }
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            data['custom_id'] = custom_id
        
        return data
    
    
    def __repr__(self):
        """Returns the message component's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def copy(self):
        """
        Copies the component.
        
        Returns
        -------
        new : ``ComponentBase``
        """
        return self
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            Keyword parameters referencing attributes.
        """
        if kwargs:
            raise TypeError(f'Unused or unsettable attributes: {kwargs!r}.')
        
        return self
    
    
    def __eq__(self, other):
        """Returns Whether the two component are equal."""
        if type(other) is not type(self):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the component's hash value."""
        return self.type.value
    
    
    def _iter_components(self):
        """
        Iterates over the sub-components recursively of the component including itself.
        
        This method is a generator.
        
        Yields
        ------
        component : ``ComponentBase``
        """
        yield self
        return
    
    
    def _replace_direct_sub_components(self, relation):
        """
        Replaces the sub components of the component with the given relation.
        
        Parameters
        ----------
        relation : `dict` of (``ComponentBase``, ``ComponentBase``) items
            Relation to replace each component with.
        """
        pass
    
    
    def _iter_direct_sub_components(self):
        """
        Iterates over the sub-components of the component.
        
        This method is a generator.
        
        Yields
        ------
        component : ``ComponentBase``
        """
        return
        yield



"""

Hata is an asynchronous Discord API wrapper built on top of scarletio. It is designed to be easy to use, with also
providing rich API offering everything what an advanced developer might need.

Named after Hata no Kokoro from Touhou Project.

Why hata
--------

- Multiple simultaneous clients

    Hata can run multiple clients from the same instance without sacrificing performance.

- Performant
    
    Fast concurrent code using async/await syntax, cache control, PyPy support and more!

- Newest API features
    
    Whatever Discord decides to release/update/break Hata will support it natively in no time!

- 100% Python

    Completely relies on Python! Easy to read, easy to understand, easy to code.

Usage
-----

The following example answers on `ping` message.

```py
from hata import Client, wait_for_interruption

Nue = Client('TOKEN')

@Nue.events
async def ready(client):
    print(f'{client:f} logged in.')

@Nue.events
async def message_create(client, message):
    if message.author.is_bot:
        return
    
    if message.content == 'ping':
        await client.message_create(message.channel, 'pong')

Nue.start()

wait_for_interruption()
```

An improved example using the `commands` extension to handle common use cases.

```py
from hata import Client, wait_for_interruption

Saki = Client('TOKEN', extensions='commands_v2', prefix='s!')

@Saki.events
async def ready(client):
    print(f'{client:f} logged in.')

@Saki.commands
async def ping(client, message):
    return 'pong'

Saki.start()

wait_for_interruption()
```

Or use slash commands!

```py
from hata import Client, Guild, wait_for_interruption

GUILD = Guild.precreate(guild_id)

Seija = Client('TOKEN', extensions='slash')

@Seija.events
async def ready(client):
    print(f'{client:f} logged in.')

@Seija.interactions(guild=GUILD)
async def ping():
    \"\"\"ping-pong\"\"\"
    return 'pong'

Seija.start()

wait_for_interruption()
```

> Note: You need to restart your client, or the slash command wont show up. If there are more than 50 integrations
> (bots) in a guild, some of the (integrations) bots wont be able to use slash commands. This is currently a Discord
> limitation.

Hata leaves the main thread free, `client.start()` blocks it only till the client logs in (or fails it), although you
can still use the `start_clients()` function, what as it says, starts up all the non-running clients parallelly.

Sometimes leaving the main thread might cause problems when trying to shut down the bot(s). At this case, you might
want to use `wait_for_interruption()`, which disconnects the clients gracefully and closes the event loop on keyboard
interrupt.

We got some tutorials on `github:https://github.com/HuyaneMatsu/hata/tree/master/docs` as well, please check them too!
"""
__version__ = '1.3.3'

from .utils.module_deprecation import get_deprecation_function

from .discord import *
from .ext import *
from .utils import *

from .env import *

__all__ = (
    *discord.__all__,
    *ext.__all__,
    
    *utils.__all__,
    *env.__all__,
)

from scarletio import check_satisfaction
check_satisfaction()

__getattr__ = get_deprecation_function()



__all__ = ('ChannelMetadataBase',)

from re import I as re_ignore_case, compile as re_compile, escape as re_escape

from scarletio import RichAttributeErrorBaseType, include

from ...bases import Icon, IconType
from ...permission import Permission
from ...permission.permission import PERMISSION_NONE
from ...utils import id_to_datetime

from ..constants import AUTO_ARCHIVE_DEFAULT
from ..flags import ChannelFlag
from ..preinstanced import VideoQualityMode


Client = include('Client')

CHANNEL_DEFAULT_ATTRIBUTES = {
    'archived': False,
    'archived_at': None,
    'auto_archive_after': AUTO_ARCHIVE_DEFAULT,
    'bitrate': 0,
    'default_auto_archive_after': AUTO_ARCHIVE_DEFAULT,
    'flags': ChannelFlag(0),
    'icon': Icon(IconType.none, 0),
    'invitable': True,
    'nsfw': False,
    'open': False,
    'owner_id': 0,
    'parent_id': 0,
    'permission_overwrites': None,
    'position': 0,
    'region': None,
    'slowmode': 0,
    'thread_users': None,
    'topic': None,
    'user_limit': 0,
    'video_quality_mode': VideoQualityMode.none,
}


class ChannelMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for channel metadatas.
    
    Class Attributes
    ----------------
    type : `int` = `-1`
        The channel's type.
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()
    
    type = -1
    order_group = 0
    
    def __new__(cls, data):
        """
        Creating channel metadatas.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data receive from Discord.
        """
        self = object.__new__(cls)
        
        self._update_attributes(data)
        
        return self
    
    
    def __repr__(self):
        """Returns the channel metadata's representation."""
        return f'<{self.__class__.__name__} type={self.type}>'
    
    
    def __eq__(self, other):
        """Returns whether the two channel metadatas are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._compare_attributes_to(other)
    
    
    def _compare_attributes_to(self, other):
        """
        Compares the channel metadata's attributes to the other one's.
        
        Parameters
        ----------
        other : ``ChannelMetadataBase``
            The other channel metadata to compare self to.
            
            > Must have the same type as self.
        
        Returns
        -------
        equal : `bool`
            Whether both metadata has the same attributes.
        """
        return True
    
    
    def _created(self, channel_entity, client):
        """
        Called when the channel entity is initialized.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The parent channel entity.
        client : `None`, ``Client``
            The client who received the channel payload.
        """
        pass
    
    
    def _to_data(cls):
        """
        Serialises the channel metadata to json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) Items
        """
        return {}
    

    def _update_attributes(self, data):
        """
        Updates the channel metadata with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        """
        pass
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the channel metadata and returns it's overwritten attributes as a `dict` with a
        `attribute-name` - `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
            
            Might contain the following items:
            
            +-------------------------------+-----------------------------------------------------------+
            | Keys                          | Values                                                    |
            +===============================+===========================================================+
            | archived                      | `bool`                                                    |
            +-------------------------------+-----------------------------------------------------------+
            | archived_at                   | `None`, `datetime`                                        |
            +-------------------------------+-----------------------------------------------------------+
            | auto_archive_after            | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | bitrate                       | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | default_auto_archive_after    | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | flags                         | ``ChannelFlag``                                           |
            +-------------------------------+-----------------------------------------------------------+
            | icon                          | ``Icon``                                                  |
            +-------------------------------+-----------------------------------------------------------+
            | invitable                     | `bool`                                                    |
            +-------------------------------+-----------------------------------------------------------+
            | name                          | `str`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | nsfw                          | `bool`                                                    |
            +-------------------------------+-----------------------------------------------------------+
            | open                          | `bool`                                                    |
            +-------------------------------+-----------------------------------------------------------+
            | owner_id                      | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | parent_id                     | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | permission_overwrites         | `None`, `dict` of (`int`, ``PermissionOverwrite``) items  |
            +-------------------------------+-----------------------------------------------------------+
            | position                      | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | region                        | `None`, ``VoiceRegion``                                   |
            +-------------------------------+-----------------------------------------------------------+
            | slowmode                      | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | topic                         | `None`, `str`                                             |
            +-------------------------------+-----------------------------------------------------------+
            | type                          | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | user_limit                    | `int`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | video_quality_mode            | ``VideoQualityMode``                                      |
            +-------------------------------+-----------------------------------------------------------+
        """
        return {}
    
    
    @classmethod
    def _from_partial_data(cls, data):
        """
        Creates a channel metadata from the given partial data. Called by ``Channel._from_partial_data`` when a
        new partial channel is needed to be created.
        
        Parameters
        ----------
        data : `None`, `dict` of (`str`, `Any`) items
            Partial channel data.
        
        Returns
        -------
        self : ``ChannelMetadataBase``
        """
        return cls._create_empty()
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates a partial channel from the given parameters.
        
        Returns
        -------
        self : ``ChannelMetadataBase``
        """
        return object.__new__(cls)
    
    
    def _delete(self, channel_entity, client):
        """
        Called when the channel is deleted.
        
        Removes the channel's references.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        client : `None`, ``Client``
            The parent client entity,
        """
        pass
    
    
    def _get_users(self, channel_entity):
        """
        The users who can see this channel.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return []
    
    
    def _iter_users(self, channel_entity):
        """
        Iterates over the users who can see the channel.
        
        This method is a generator.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        yield from self.users
    
    
    def _get_clients(self, channel_entity):
        """
        Helper class to get the channel's clients.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        
        Returns
        -------
        clients : `list` of ``Client``
        """
        clients = []
        for user in self._get_users(channel_entity):
            if isinstance(user, Client):
                clients.append(user)
        
        return clients
    
    
    def _get_user(self, channel_entity, name, default):
        """
        Tries to find the a user with the given name at the channel. Returns the first matched one.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        name : `str`
            The name to search for.
        default : `Any`
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase``, `default`
        """
        if (not 1 < len(name) < 38):
            return default
        
        users = self._get_users(channel_entity)
        
        if len(name) > 6 and name[-5] == '#':
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if user.discriminator == discriminator and user.name == name_:
                        return user
        
        if len(name) > 32:
            return default
        
        for user in users:
            if user.name == name:
                return user
        
        return default
    
    
    def _get_user_like(self, channel_entity, name, default):
        """
        Searches a user, who's name or nick starts with the given string and returns the first find.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        name : `str`
            The name to search for.
        default : `Any`
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase``, `default`
        """
        name_length = len(name)
        if name_length < 1:
            return default
        
        users = self._get_users(channel_entity)
        
        if name_length > 6:
            if name_length > 37:
                return default
            
            if name[-5] == '#':
                try:
                    discriminator = int(name[-4:])
                except ValueError:
                    pass
                else:
                    stripped_name = name[:-5]
                    for user in users:
                        if user.discriminator == discriminator and user.name == stripped_name:
                            return user
        
        if name_length > 32:
            return default
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        for user in users:
            if pattern.search(user.name) is None:
                continue
            
            return user
        
        return default
    
    
    def _get_users_like(self, channel_entity, name):
        """
        Searches the users, who's name or nick starts with the given string.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        name : `str`
            The name to search for.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        result = []
        
        name_length = len(name)
        if name_length < 1:
            return result
        
        users = self._get_users(channel_entity)
        
        if name_length > 6:
            if name_length > 37:
                return result
            
            if name[-5] == '#':
                try:
                    discriminator = int(name[-4:])
                except ValueError:
                    pass
                else:
                    stripped_name = name[:-5]
                    for user in users:
                        if user.discriminator == discriminator and user.name == stripped_name:
                            result.append(user)
                            break
        
        if name_length > 32:
            return result
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        for user in users:
            if pattern.search(user.name) is None:
                continue
            
            result.append(user)
        
        return result
    
    
    @property
    def name(self):
        """
        Returns the channel's name.
        
        Returns
        -------
        name : `str`
        """
        return 'channel'
    
    
    def _get_processed_name(self):
        """
        Returns the channel's name.
        
        Returns
        -------
        name : `str`
        """
        return self.name
    
    
    def _get_display_name(self):
        """
        Returns channel's display name.
        
        Returns
        -------
        display_name : `str`
        """
        return self.name
    

    def _get_permissions_for(self, channel_entity, user):
        """
        Returns the permissions for the given user at the channel.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
            
        
        See Also
        --------
        ``.cached_permissions_for`` : Cached permission calculator.
        """
        return PERMISSION_NONE
    
    
    def _get_cached_permissions_for(self, channel_entity, user):
        """
        Returns the permissions for the given user at the channel. If the user's permissions are not cached, calculates
        and stores them first.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        """
        return self._get_permissions_for(channel_entity, user)
    
    
    def _get_permissions_for_roles(self, channel_entity, roles):
        """
        Returns the channel permissions of an imaginary user who would have the listed roles.
        
        Parameters
        ----------
        roles : `tuple` of ``Role``
            The roles to calculate final permissions from.
        
        Returns
        -------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        permissions : ``Permission``
            The calculated permissions.
        """
        return PERMISSION_NONE
    
    
    def __dir__(self):
        """Returns the attributes of the channel."""
        return sorted(set(object.__dir__(self))|set(CHANNEL_DEFAULT_ATTRIBUTES.keys()))
    
    
    def __getattr__(self, attribute_name):
        """Returns the channel metadata's attribute if found."""
        try:
            return CHANNEL_DEFAULT_ATTRIBUTES[attribute_name]
        except KeyError:
            pass
        
        return RichAttributeErrorBaseType.__getattr__(self, attribute_name)
    
    
    @classmethod
    def _precreate(cls, keyword_parameters):
        """
        Precreates the channel metadata. Each channel type exhaust it's own specific attributes from
        `keyword_parameters`.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters passed to ``Channel.precreate``
        
        Returns
        -------
        self : ``ChannelMetadataBase``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        return cls._create_empty()
    
    
    def _get_created_at(self, channel_entity):
        """
        Returns when the channel was created.
        
        Returns
        -------
        created_at : `datetime`
        """
        return id_to_datetime(channel_entity.id)
    
    
    def _invalidate_permission_cache(self):
        """
        Invalidates the cached permissions of the channel.
        
        This method is only applicable for channel types with permission cache.
        """
        pass



__all__ = ()

COMPONENT_SUB_COMPONENT_LIMIT = 5
COMPONENT_LABEL_LENGTH_MAX = 80
COMPONENT_CUSTOM_ID_LENGTH_MAX = 100
COMPONENT_OPTION_LENGTH_MIN = 1
COMPONENT_OPTION_LENGTH_MAX = 25
COMPONENT_OPTION_MIN_VALUES_MIN = 0
COMPONENT_OPTION_MIN_VALUES_MAX = 15
COMPONENT_OPTION_MAX_VALUES_MIN = 1
COMPONENT_OPTION_MAX_VALUES_MAX = 25
COMPONENT_MAX_LENGTH_MIN = 0
COMPONENT_MAX_LENGTH_MAX = 4000
COMPONENT_MIN_LENGTH_MIN = 0
COMPONENT_MIN_LENGTH_MAX = 4000
COMPONENT_TITLE_LENGTH_MIN = 1
COMPONENT_TITLE_LENGTH_MAX = 45



__all__ = ()

import warnings

from scarletio import Compound

from ...bases import maybe_snowflake
from ...channel import Channel
from ...core import GUILDS
from ...guild import Guild, create_partial_guild_from_id
from ...http import DiscordHTTPClient
from ...role import Role
from ...user import ClientUserBase, User
from ...utils import datetime_to_timestamp

from ..request_helpers import (
    get_guild_and_id, get_guild_id, get_channel_guild_id_and_id, get_role_guild_id_and_id, get_user_and_id, get_user_id,
    validate_timeout_duration
)


def _assert__user_guild_profile_edit__nick(nick):
    """
    Asserts the `nick` parameter of ``Client.user_guild_profile_edit`` method.
    
    Parameters
    ----------
    nick : `Ellipsis`, `None`, `int`
        The new nick of the user.
    
    Raises
    ------
    AssertionError
        - If `nick` was not given neither as `None`, `str`.
        - If `nick` length is out of the expected range [0:32].
    """
    if nick is ...:
        pass
    
    elif nick is None:
        pass
    
    elif isinstance(nick, str):
        nick_length = len(nick)
        if nick_length > 32:
            raise AssertionError(
                f'`nick` length can be in range [0:32], got {nick_length}; {nick!r}.'
            )
    
    else:
        raise AssertionError(
            f'`nick` can be `None`, `str`, got {nick.__class__.__name__}; {nick!r}.'
        )
    
    return True


def _assert__user_guild_profile_edit__deaf(deaf):
    """
    Asserts the `deaf` parameter of ``Client.user_guild_profile_edit`` method.
    
    Parameters
    ----------
    deaf : `Ellipsis`, `bool`
        Whether the user should be deafen at the voice channels.
    
    Raises
    ------
    AssertionError
        - If `deaf` was not given as `bool`.
    """
    if (deaf is not ...) and (not isinstance(deaf, bool)):
        raise TypeError(
            f'`deaf` can be `bool`, got {deaf.__class__.__name__}; {deaf!r}'
        )
    
    return True


def _assert__user_guild_profile_edit__mute(mute):
    """
    Asserts the `mute` parameter of ``Client.user_guild_profile_edit`` method.
    
    Parameters
    ----------
    mute : `Ellipsis`, `bool`
        Whether the user should be muted at the voice channels.
    
    Raises
    ------
    AssertionError
        - If `mute` was not given as `bool`.
    """
    if (mute is not ...) and (not isinstance(mute, bool)):
        raise TypeError(
            f'`mute` can be `bool`, got {mute.__class__.__name__}; {mute!r}'
        )
    
    return True


def _assert__guild_user_search__query(query):
    """
    Asserts the `query` parameter of ``Client.guild_user_search`` method.
    
    Parameters
    ----------
    query : `str`
        The query string with what the user's name or nick should start.
    
    Raises
    ------
    AssertionError
        - If `query` was not given as `str`.
        - If `query`'s length is out of the expected range [1:32].
    """
    if not isinstance(query, str):
        raise TypeError(
            f'`query` can be `str`, got {query.__class__.__name__}; {query!r}'
        )
    
    query_length = len(query)
    if query_length < 1 or query_length > 32:
        raise AssertionError(
            f'`query` length can be in range [1:32], got {query_length!r}; {query!r}.'
        )
    
    return True


def _assert__guild_user_search__limit(limit):
    """
    Asserts the `limit` parameter of ``Client.guild_user_search`` method.
    
    Parameters
    ----------
    limit : `int`
        The maximal amount of users to return. Can be in range [1:1000].
    
    Raises
    ------
    AssertionError
        - If `limit` was not given as `int`.
    """
    if not isinstance(limit, int):
        raise TypeError(
            f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}'
        )
    
    return True


class ClientCompoundUserEndpoints(Compound):
    
    http : DiscordHTTPClient
    id : int
    
    async def user_guild_profile_edit(
        self, guild, user, *, nick=..., deaf=..., mute=..., voice_channel=..., roles=..., timeout_duration=...,
        timed_out_until=..., reason=None
    ):
        """
        Edits the user's guild profile at the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            Where the user will be edited.
        
        user : ``ClientUserBase``, `int`
            The user to edit, or their identifier.
        
        nick : `None`, `str`, Optional (Keyword only)
            The new nick of the user. You can remove the current one by passing it as `None`, an empty string.
        
        deaf : `bool`, Optional (Keyword only)
            Whether the user should be deafen at the voice channels.
        
        mute : `bool`, Optional (Keyword only)
            Whether the user should be muted at the voice channels.
        
        voice_channel : `None`, ``Channel``, `int` , Optional (Keyword only)
            Moves the user to the given voice channel. Only applicable if the user is already at a voice channel.
            
            Pass it as `None` to disconnect the user from it's voice channel.
        
        roles : `None`, `iterable` of (``Role``, `int`), Optional (Keyword only)
            The new roles of the user. Give it as `None` to remove all of the user's roles.
        
        timeout_duration: `None`, `int`, `float`, `timedelta`, `datetime`, Optional (Keyword only)
            The timeout duration of the user in seconds.
            
            Pass it as `None` or as a non-positive duration to remove it.
            
            The max allowed value equals to 28 days.
            
        timed_out_until : `None`, `datetime`, Optional (Keyword only)
            Till when the user is timed out. Pass it as `None` to remove it.
            
            > Deprecated and will be removed in 2022 december. Please use `timeout_duration` instead.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        user, user_id = get_user_and_id(user)
        
        assert _assert__user_guild_profile_edit__nick(nick)
        assert _assert__user_guild_profile_edit__deaf(deaf)
        assert _assert__user_guild_profile_edit__mute(mute)
        
        data = {}
        if (nick is not ...):
            if (nick is not None) and (not nick):
                nick = None
            
            if (guild is not None) and (user is not None) and guild.partial:
                try:
                    guild_profile = user.guild_profiles[guild.id]
                except KeyError:
                    should_edit_nick = True
                else:
                    if guild_profile.nick == nick:
                        should_edit_nick = False
                    else:
                        should_edit_nick = True
            else:
                should_edit_nick = True
            
            if should_edit_nick:
                if self.id == user_id:
                    await self.http.client_guild_profile_edit(guild_id, {'nick': nick}, reason)
                else:
                    data['nick'] = nick
                    
        if (deaf is not ...):
            data['deaf'] = deaf
        
        if (mute is not ...):
            data['mute'] = mute
        
        if (voice_channel is not ...):
            while True:
                if voice_channel is None:
                    voice_channel_id = None
                    break
                
                elif isinstance(voice_channel, Channel):
                    if voice_channel.is_in_group_guild_connectable() or voice_channel.partial:
                        voice_channel_id = voice_channel.id
                        break
                
                else:
                    voice_channel_id = maybe_snowflake(voice_channel)
                    if voice_channel_id is not None:
                        break
                
                raise TypeError(
                    f'`voice_channel` can be `None`, any guild connectable channel, `int`, got '
                    f'{voice_channel.__class__.__name__}; {voice_channel!r}.'
                )
            
            data['channel_id'] = voice_channel_id
        
        
        if (roles is not ...):
            role_ids = set()
            if (roles is not None):
                iterator = getattr(roles, '__iter__', None)
                if iterator is None:
                    raise TypeError(
                        f'`roles` can be `None`, `iterable`, got {roles.__class__.__name__}; {roles!r}.'
                    )
                
                for role in roles:
                    
                    if isinstance(role, Role):
                        role_id = role.id
                    else:
                        role_id = maybe_snowflake(role)
                        if role_id is None:
                            raise TypeError(
                                f'`roles` can contain `{Role.__name__}`, `int` elements, got '
                                f'{role.__class__.__name__}; {role!r}; roles={roles!r}.'
                            )
                    
                    role_ids.add(role_id)
            
            data['roles'] = role_ids
        
        
        if (timed_out_until is not ...):
            warnings.warn(
                (
                    f'`timed_out_until` parameter of `{self.__class__.__name__}.user_guild_profile_edit` is deprecated'
                    f'and will be removed in 2022 December. Please use `timeout_duration` instead.'
                ),
                FutureWarning,
            )
            timeout_duration = timed_out_until
        
        
        if (timeout_duration is not ...):
            timeout_ends_at = validate_timeout_duration(timeout_duration)
            
            if (timeout_ends_at is None):
                timed_out_until_raw = None
            else:
                timed_out_until_raw = datetime_to_timestamp(timeout_ends_at)
            data['communication_disabled_until'] = timed_out_until_raw
        
        await self.http.user_guild_profile_edit(guild_id, user_id, data, reason)
    
    
    async def user_role_add(self, user, role, *, reason=None):
        """
        Adds the role on the user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user who will get the role.
        role : ``Role``, `tuple` (`int`, `int`)
            The role to add on the user.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `role` was not given neither as ``Role`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        snowflake_pair = get_role_guild_id_and_id(role)
        if snowflake_pair is None:
            return
        guild_id, role_id = snowflake_pair
        user_id = get_user_id(user)
        
        await self.http.user_role_add(guild_id, user_id, role_id, reason)
    
    
    async def user_role_delete(self, user, role, *, reason=None):
        """
        Deletes the role from the user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user from who the role will be removed.
        role : ``Role``, `tuple` (`int`, `int`)
            The role to remove from the user.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `role` was not given neither as ``Role`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        snowflake_pair = get_role_guild_id_and_id(role)
        if snowflake_pair is None:
            return
        guild_id, role_id = snowflake_pair
        user_id = get_user_id(user)
        
        await self.http.user_role_delete(guild_id, user_id, role_id, reason)
    
    
    async def user_voice_move(self, user, channel):
        """
        Moves the user to the given voice channel. The user must be in a voice channel at the respective guild already.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user to move.
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel where the user will be moved.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `channel` was not given neither as ``Channel`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_guild_connectable)
        if not guild_id:
            return
        
        user_id = get_user_id(user)
       
        await self.http.user_move(guild_id, user_id, {'channel_id': channel_id})
    
    
    async def user_voice_move_to_speakers(self, user, channel):
        """
        Moves the user to the speakers inside of a stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user to move.
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel where the user will be moved.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `channel` was not given neither as ``Channel`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        if not guild_id:
            return
        
        user_id = get_user_id(user)
       
        data = {
            'suppress' : False,
            'channel_id': channel_id,
        }
        
        await self.http.voice_state_user_edit(guild_id, user_id, data)
    
    
    async def user_voice_move_to_audience(self, user, channel):
        """
        Moves the user to the audience inside of a stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user to move.
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel where the user will be moved.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `channel` was not given neither as ``Channel`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        if not guild_id:
            return
        
        user_id = get_user_id(user)
       
        data = {
            'suppress': True,
            'channel_id': channel_id,
        }
        
        await self.http.voice_state_user_edit(guild_id, user_id, data)
    
    
    async def user_voice_kick(self, user, guild):
        """
        Kicks the user from the guild's voice channels. The user must be in a voice channel at the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user who will be kicked from the voice channel.
        guild : ``Guild``, `int`
            The guild from what's voice channel the user will be kicked.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        user_id = get_user_id(user)
        guild_id = get_guild_id(guild)
        
        await self.http.user_move(guild_id, user_id, {'channel_id': None})
    
    
    async def user_get(self, user, *, force_update=False):
        """
        Gets an user by it's id. If the user is already loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ``ClientUserBase``, `int`
            The user, who will be requested.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the user should be requested even if it supposed to be up to date.
        
        Returns
        -------
        user : ``ClientUserBase``
        
        Raises
        ------
        TypeError
            If `user` was not given neither as ``ClientUserBase``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Raises
        ------
        TypeError
            If `user` was not given as ``ClientUserBase``, nor as `int`.
        """
        user, user_id = get_user_and_id(user)
        
        # a goto to check whether we should force update the user.
        while True:
            if force_update:
                break
            
            if user is None:
                break
            
            for guild_id in user.guild_profiles.keys():
                try:
                    guild = GUILDS[guild_id]
                except KeyError:
                    continue
                
                if not guild.partial:
                    return user
            
            break
        
        data = await self.http.user_get(user_id)
        return User._create_and_update(data)
    
    
    async def guild_user_get(self, guild, user):
        """
        Gets an user and it's profile at a guild. The user must be the member of the guild. If the user is already
        loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, where the user is.
        user : ```ClientUserBase``, `int`
            The user's id, who will be requested.
        
        Returns
        -------
        user : ``ClientUserBase``
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        user_id = get_user_id(user)
        guild, guild_id = get_guild_and_id(guild)
        
        data = await self.http.guild_user_get(guild_id, user_id)
        
        if guild is None:
            guild = create_partial_guild_from_id(guild_id)
        
        return User._create_and_update(data, guild)
    
    
    async def guild_user_search(self, guild, query, limit=1):
        """
        Gets an user and it's profile at a guild by it's name. If the users are already loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild, where the user is.
        
        query : `name`
            The query string with what the user's name or nick should start.
        
        limit : `int` = `1`, Optional
            The maximal amount of users to return. Can be in range [1:1000].
            
            When passed as non-positive value, will default to the maximal allowed amount.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `str`.
            - If `limit` is out fo expected range [1:1000].
        """
        guild, guild_id = get_guild_and_id(guild)
        
        assert _assert__guild_user_search__query(query)
        assert _assert__guild_user_search__limit(limit)
        
        if limit <= 0 or limit > 1000:
            limit = 1000
        
        data = {'query': query}
        
        if limit != 1:
            data['limit'] = limit
        
        data = await self.http.guild_user_search(guild_id, data)
        
        if guild is None:
            guild = create_partial_guild_from_id(guild_id)
        
        return [User._create_and_update(user_data, guild) for user_data in data]



from .add_default_plugin_variables_ import *
from .clear_default_plugin_variables_ import *
from .get_plugin_ import *
from .get_plugin_like_ import *
from .get_plugins_like_ import *
from .import_plugin_ import *
from .load_all_plugin_ import *
from .load_plugin_ import *
from .register_and_load_plugin_ import *
from .register_plugin_ import *
from .reload_all_plugin_ import *
from .reload_plugin_ import *
from .remove_default_plugin_variables_ import *
from .require_ import *
from .unload_all_plugin_ import *
from .unload_plugin_ import *


__all__ = (
    *add_default_plugin_variables_.__all__,
    *clear_default_plugin_variables_.__all__,
    *get_plugin_.__all__,
    *get_plugin_like_.__all__,
    *get_plugins_like_.__all__,
    *import_plugin_.__all__,
    *load_all_plugin_.__all__,
    *load_plugin_.__all__,
    *register_and_load_plugin_.__all__,
    *register_plugin_.__all__,
    *reload_all_plugin_.__all__,
    *reload_plugin_.__all__,
    *remove_default_plugin_variables_.__all__,
    *require_.__all__,
    *unload_all_plugin_.__all__,
    *unload_plugin_.__all__,
)



__all__ = ('ApplicationCommandPermissionOverwrite',)

from scarletio import RichAttributeErrorBaseType

from ...channel import Channel
from ...core import CHANNELS, ROLES
from ...role import Role, create_partial_role_from_id
from ...user import ClientUserBase, create_partial_user_from_id

from .preinstanced import ApplicationCommandPermissionOverwriteTargetType


APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER = ApplicationCommandPermissionOverwriteTargetType.user
APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE = ApplicationCommandPermissionOverwriteTargetType.role
APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL = ApplicationCommandPermissionOverwriteTargetType.channel


def _validate_application_command_target(target):
    """
    Validates the given application command's input target.
    
    Parameters
    ----------
    target : ``ClientUserBase``, ``Role``, ``Channel``, `tuple` ((``ClientUserBase``, ``Role``, \
            ``Channel``, `str` (`'Role'`, `'role'`, `'User'`, `'user'`, `'Channel'`, `'channel'`)), `int`)
        The target entity of the application command permission overwrite.
    
    Returns
    -------
    target_type : ``ApplicationCommandPermissionOverwriteTargetType``
        The target entity's type.
    
    target_id : `int`
        The represented entity's identifier.
    
    Raises
    ------
    TypeError
        - If `target` was not given as any of the expected types & values.
    """
    # GOTO
    while True:
        if isinstance(target, Role):
            target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
            target_id = target.id
            target_lookup_failed = False
            break
        
        if isinstance(target, ClientUserBase):
            target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
            target_id = target.id
            target_lookup_failed = False
            break
        
        if isinstance(target, Channel):
            target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL
            target_id = target.id
            target_lookup_failed = False
            break
        
        if isinstance(target, tuple) and len(target) == 2:
            target_maybe, target_id_maybe = target
            
            if isinstance(target_maybe, type):
                if issubclass(target_maybe, Role):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
                elif issubclass(target_maybe, ClientUserBase):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
                elif issubclass(target_maybe, Channel):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL
                else:
                    target_lookup_failed = True
                    break
            
            elif isinstance(target_maybe, str):
                if target_maybe in ('Role', 'role'):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
                elif target_maybe in ('User', 'user'):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
                elif target_maybe in ('Channel', 'channel'):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL
                else:
                    target_lookup_failed = True
                    break
            
            else:
                target_lookup_failed = True
                break
            
            if type(target_id_maybe) is int:
                target_id = target_id_maybe
            elif isinstance(target_id_maybe, int):
                target_id = int(target_id_maybe)
            else:
                target_lookup_failed = True
                break
            
            target_lookup_failed = False
            break
        
        target_lookup_failed = True
        break
    
    if target_lookup_failed:
        raise TypeError(
            f'`target` can be `{Role.__name__}`, `{ClientUserBase.__name__}`, `{Channel.__name__}`, '
            f'`tuple` ((`{Role.__name__}`, `{ClientUserBase.__name__}`, `{Channel.__name__}`, `str` '
            f'(`\'Role\'`, `\'role\'`, `\'User\'`, `\'user\'`, `\'Channel\'`, `\'channel\'`)), `int`), '
            f'got {target.__class__.__name__}: {target!r}.'
        )
    
    return target_type, target_id


class ApplicationCommandPermissionOverwrite(RichAttributeErrorBaseType):
    """
    Represents an application command's allow/disallow overwrite for the given entity.
    
    Attributes
    ----------
    allow : `bool`
        Whether the respective command is allowed for the represented entity.
    
    target_id : `int`
        The represented entity's identifier.
    
    target_type : ``ApplicationCommandPermissionOverwriteTargetType``
        The target entity's type.
    """
    __slots__ = ('allow', 'target_id', 'target_type')
    
    def __new__(cls, target, allow):
        """
        Creates a new ``ApplicationCommandPermission`` with the given parameters.
        
        Parameters
        ----------
        target : ``ClientUserBase``, ``Role``, ``Channel``, `tuple` ((``ClientUserBase``, ``Role``, \
                ``Channel``, `str` (`'Role'`, `'role'`, `'User'`, `'user'`, `'Channel'`, `'channel'`)), `int`)
            The target entity of the application command permission overwrite.
            
            The expected type & value might be pretty confusing, but the target was it to allow relaxing creation.
            To avoid confusing, here is a list of the expected structures:
            
            - ``Role``
            - ``ClientUserBase``
            - ``Channel``
            - `tuple` (``Role``, `int`)
            - `tuple` (``ClientUserBase``, `int`)
            - `tuple` (``Channel``, `int`)
            - `tuple` (`'Role'`, `int`)
            - `tuple` (`'role'`, `int`)
            - `tuple` (`'User'`, `int`)
            - `tuple` (`'user'`, `int`)
            - `tuple` (`'Channel'`, `int`)
            - `tuple` (`'channel'`, `int`)
        
        allow : `bool`
            Whether the respective application command should be enabled for the respective entity.
        
        Raises
        ------
        TypeError
            - If `target` was not given as any of the expected types & values.
        AssertionError
            - If `allow` was not given as `bool`.
        """
        # allow
        if __debug__:
            if not isinstance(allow, bool):
                raise AssertionError(
                    f'`allow` can be `bool`, got {allow.__class__.__name__}; {allow!r}.'
                )
        
        # target_id & target_type
        target_type, target_id = _validate_application_command_target(target)
        
        self = object.__new__(cls)
        self.allow = allow
        self.target_id = target_id
        self.target_type = target_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandPermissionOverwrite`` from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command permission overwrite data.
        
        Returns
        -------
        self : ``ApplicationCommandPermission``
            The created application command option.
        """
        # allow
        allow = data['permission']
        
        # target_id
        target_id = int(data['id'])
        
        # target_type
        target_type = ApplicationCommandPermissionOverwriteTargetType.get(data['type'])
        
        self = object.__new__(cls)
        self.allow = allow
        self.target_id = target_id
        self.target_type = target_type
        return self
    
    
    def to_data(self):
        """
        Converts the application command permission overwrite to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # allow
        data['permission'] = self.allow
        
        # target_id
        data['id'] = self.target_id
        
        # target_type
        data['type'] = self.target_type.value
        
        return data
    
    
    @property
    def target(self):
        """
        Returns the application command overwrite's target entity.
        
        Returns
        -------
        target : `None`, ``Role``, ``ClientUserBase``, ``Channel``
        """
        target_type = self.target_type
        target_id = self.target_id
        
        if target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE:
            target = create_partial_role_from_id(target_id)
        
        elif target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER:
            target = create_partial_user_from_id(target_id)
        
        elif target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL:
            target = CHANNELS.get(target_id)
        
        else:
            target = None
        
        return target
    
    
    def __repr__(self):
        """Returns the application command permission overwrite's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Primary fields: `.target_type`, `.target_id`
        repr_parts.append(' target_type=')
        repr_parts.append(self.target_type.name)
        
        repr_parts.append(', target_id=')
        repr_parts.append(repr(self.target_id))
        
        # Secondary fields: `.allow`
        repr_parts.append(', allow=')
        repr_parts.append(repr(self.allow))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application command overwrites are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # allow
        if self.allow != other.allow:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        # target_type
        if self.target_type is not other.target_type:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the application command permission overwrite's hash value."""
        hash_value = 0
        
        # allow
        hash_value ^= self.allow << 8
        
        # target_id
        hash_value ^= self.target_id
        
        # target_type
        hash_value ^= self.target_type.value
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the application command permission overwrite.
        
        Returns
        -------
        new : ``ApplicationCommandPermissionOverwrite``
        """
        new = object.__new__(type(self))
        
        # allow
        new.allow = self.allow
        
        # target_id
        new.target_id = self.target_id
        
        # target_type
        new.target_type = self.target_type
        
        return new
    
    
    def __gt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_type_value = self.target_type.value
        other_type_value = other.target_type.value
        
        if self_type_value > other_type_value:
            return True
        
        if self_type_value < other_type_value:
            return False
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_role = ROLES.get(self_target_id, None)
            other_role = ROLES.get(other_target_id, None)
            if self_role is None:
                if other_role is None:
                    return (self_target_id > other_target_id)
                else:
                    return False
            else:
                if other_role is None:
                    return True
                else:
                    return (self_role > other_role)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER.value:
            return (self.target_id > other.target_id)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_channel = CHANNELS.get(self_target_id, None)
            other_channel = CHANNELS.get(other_target_id, None)
            if self_channel is None:
                if other_channel is None:
                    return (self_target_id > other_target_id)
                else:
                    return False
            else:
                if other_channel is None:
                    return True
                else:
                    return (self_channel.name > other_channel.name)
        
        # Should not happen
        return False
    
    
    def __lt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_type_value = self.target_type.value
        other_type_value = other.target_type.value
        
        if self_type_value > other_type_value:
            return False
        
        if self_type_value < other_type_value:
            return True
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_role = ROLES.get(self_target_id, None)
            other_role = ROLES.get(other_target_id, None)
            if self_role is None:
                if other_role is None:
                    return (self_target_id < other_target_id)
                else:
                    return True
            else:
                if other_role is None:
                    return False
                else:
                    return (self_role < other_role)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER.value:
            return (self.target_id < other.target_id)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_channel = CHANNELS.get(self_target_id, None)
            other_channel = CHANNELS.get(other_target_id, None)
            if self_channel is None:
                if other_channel is None:
                    return (self_target_id < other_target_id)
                else:
                    return True
            else:
                if other_channel is None:
                    return False
                else:
                    return (self_channel.name < other_channel.name)
        
        # Should not happen
        return True
    
    
    def copy_width(self, **keyword_parameters):
        """
        Copies the application command permission overwrite with modifying the given parameters inside of it.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            The attributes to change.
        
        Other Parameters
        ----------------
        allow : `bool`
            Whether the respective application command should be enabled for the respective entity.
        
        target : ``ClientUserBase``, ``Role``, ``Channel``, `tuple` ((``ClientUserBase``, ``Role``, \
                ``Channel``, `str` (`'Role'`, `'role'`, `'User'`, `'user'`, `'Channel'`, `'channel'`)), `int`)
            The target entity of the application command permission overwrite.
        
        Returns
        -------
        new : ``ApplicationCommandPermissionOverwrite``
            the newly created application command permission overwrite.
        
        Raises
        ------
        TypeError
            - If `target` was not given as any of the expected types & values.
            - If extra parameters were given.
        AssertionError
            - If `allow` was not given as `bool`.
        """
        # allow
        try:
            allow = keyword_parameters.pop('allow')
        except KeyError:
            allow = self.allow
        else:
            if __debug__:
                if not isinstance(allow, bool):
                    raise AssertionError(
                        f'`allow` can be `bool`, got {allow.__class__.__name__}; {allow!r}.'
                    )
        
        # target_id & target_type
        try:
            target = keyword_parameters.pop('target')
        except KeyError:
            target_id = self.target_id
            target_type = self.target_type
        else:
            target_type, target_id = _validate_application_command_target(target)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        new = object.__new__(type(self))
        new.allow = allow
        new.target_id = target_id
        new.target_type = target_type
        return new



__all__ = ('reload_plugin',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.reload)
def reload_plugin(name, *, blocking=True, deep=True):
    return PLUGIN_LOADER.reload(name, blocking=blocking, deep=deep)



from .bots_query import *
from .client import *
from .constants import *
from .exceptions import *
from .rate_limit_handling import *
from .types import *

__all__ = (
    'setup_ext_top_gg',
    *bots_query.__all__,
    *client.__all__,
    *constants.__all__,
    *exceptions.__all__,
    *rate_limit_handling.__all__,
    *types.__all__,
)

# Implement setup function

from .client import _start_auto_post, _stop_auto_post

def setup_ext_top_gg(client, *args, **kwargs):
    """
    Setups the top.gg extension on client.
    
    Note, that this function can be called on a client only once.
    
    Parameters
    ----------
    client : ``Client``
        The client to setup the extension on.

    **kwargs : Keyword parameters
        Additional keyword parameter to be passed to the created ``Slasher``.
    **kwargs : Keyword parameters
        Additional keyword parameter to be passed to the created ``Slasher``.
    
    Other Parameters
    ----------------
    top_gg_token : `str`
        Top.gg api token.
    auto_post_bot_stats : `bool`, Optional
        Whether auto post should be started as the client launches up.
        
        Defaults to `True`.
    raise_on_top_gg_global_rate_limit : `bool`, Optional
        Whether ``TopGGGloballyRateLimited`` should be raised when the client gets globally rate limited.
        
        Defaults to `False`.
    
    Returns
    -------
    top_gg_client : ``Slasher``
        Top gg client added to the client.
    
    Raises
    ------
    RuntimeError
        If the client has an attribute set what the top.gg client would use.
    TypeError
        - If `client` is not ``Client``.
        - If `top_gg_token` is not `str`.
        - If `auto_post_bot_stats` is not `bool`.
        - If `raise_on_top_gg_global_rate_limit` is not `bool` isinstance.
    """
    if hasattr(client, 'top_gg'):
        raise RuntimeError(f'The client already has `top_gg_client` attribute; got {client!r}.')
    
    top_gg_client = TopGGClient(client, *args, **kwargs)
    
    client.top_gg = top_gg_client
    client.events(_start_auto_post, name='launch')
    client.events(_stop_auto_post, name='shutdown')
    
    return top_gg_client

# Register extension

from .. import register_library_extension, register_setup_function

register_library_extension('HuyaneMatsu.top_gg')

register_setup_function(
    'HuyaneMatsu.top_gg',
    setup_ext_top_gg,
    (
        'top_gg_token',
    ),
    (
        'auto_post_bot_stats',
        'raise_on_top_gg_global_rate_limit',
    ),
)



__all__ = (
    'AutoModerationRuleTriggerMetadata', 'KeywordPresetTriggerMetadata', 'KeywordTriggerMetadata',
    'MentionSpamTriggerMetadata'
)

from scarletio import RichAttributeErrorBaseType, copy_docs, include

from .constants import AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX


AutoModerationKeywordPresetType = include('AutoModerationKeywordPresetType')


def try_get_auto_moderation_trigger_metadata_type_from_data(data):
    """
    Tries to detect what type of auto moderation trigger metadata the given data is.
    
    Parameters
    ----------
    data : `dict` of (`str`, `str`) items
        Auto moderation trigger metadata data.
    
    Returns
    -------
    metadata_type : `None`, `type`
    """
    if 'keyword_filter' in data:
        metadata_type = KeywordTriggerMetadata
    elif 'presets' in data:
        metadata_type = KeywordPresetTriggerMetadata
    else:
        metadata_type = None
    
    return metadata_type


class AutoModerationRuleTriggerMetadata(RichAttributeErrorBaseType):
    """
    Base class for ``AutoModerationRule``'s trigger metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new trigger metadata instance.
        """
        return object.__new__(cls)
        
    def __repr__(self):
        """Returns the trigger metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new auto moderation rule trigger metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Auto moderation rule trigger metadata payload.
        
        Returns
        -------
        self : ``ScheduledEventEntityMetadata``
        """
        return object.__new__(cls)
    
    
    def to_data(self):
        """
        Converts the trigger metadata to json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __eq__(self, other):
        """Returns whether the two trigger metadatas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the trigger metadata's hash value."""
        return 0
    
    
    def copy(self):
        """
        Copies the trigger metadata.
        
        Returns
        -------
        new : ``AutoModerationRuleTriggerMetadata``
        """
        return object.__new__(type(self))


class KeywordTriggerMetadata(AutoModerationRuleTriggerMetadata):
    """
    Keyword trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    keywords : `None`, `tuple` of `str`
        Substrings which will be searched for in content.
    
    Matching Strategies
    -------------------
    Use the `*` wildcard symbol at the beginning and end of a keyword to define how the keyword should be matched.
    All keywords are not case sensitive.
    
    - `Prefix` - word must start with the keyword
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | cat*      | catch, Catapult, CAttLE               |
        +-----------+---------------------------------------+
        | tra*      | train, trade, TRAditional             |
        +-----------+---------------------------------------+
        | the mat*  | the matrix                            |
        +-----------+---------------------------------------+
    
    
    - `Suffix` - word must end with the keyword
    
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | *cat      | wildcat, copyCat                      |
        +-----------+---------------------------------------+
        | *tra      | extra, ultra, orchesTRA               |
        +-----------+---------------------------------------+
        | *the mat  | breathe mat                           |
        +-----------+---------------------------------------+
    
    
    - `Anywhere` - keyword can appear anywhere in the content
    
        +-----------+---------------------------------------+
        | Keyword     | Matches                             |
        +===========+=======================================+
        | *cat*       | location, eduCation                 |
        +-----------+---------------------------------------+
        | *tra*       | abstracted, outrage                 |
        +-----------+---------------------------------------+
        | *the mat*   | breathe matter                      |
        +-----------+---------------------------------------+
    
    
    - `Whole Word` - keyword is a full word or phrase and must be surrounded by whitespace at the beginning and end
    
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | cat       | cat                                   |
        +-----------+---------------------------------------+
        | train     | train                                 |
        +-----------+---------------------------------------+
        | the mat   | the mat                               |
        +-----------+---------------------------------------+
    """
    __slots__ = ('keywords',)
    
    def __new__(cls, keywords):
        """
        Creates a new keyword trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        keywords : `None`, `str`, `iterable` of `str`
            Substrings which will be searched for in content.
               
        Raises
        ------
        TypeError
            - If `keywords` type is incorrect.
            - If a keyword's type is incorrect.
        """
        if keywords is None:
            processed_keywords = None
        
        elif isinstance(keywords, str):
            processed_keywords = (keywords, )
        
        else:
            iterator = getattr(type(keywords), '__iter__', None)
            if iterator is None:
                raise TypeError(
                    f'`keywords` can be `None`, `str`, `iterable` of `str`, '
                    f'got {keywords.__class__.__name__}; {keywords!r}.'
                )
            
            processed_keywords = None
            
            for keyword in iterator(keywords):
                if not isinstance(keyword, str):
                    raise TypeError(
                        f'`keywords` can contain `str` elements, got {keyword.__class__.__name__}; {keyword!r}; '
                        f'keywords={keywords!r}'
                    )
                
                if processed_keywords is None:
                    processed_keywords = set()
                
                processed_keywords.add(keyword)
            
            if (processed_keywords is not None):
                processed_keywords = tuple(sorted(processed_keywords))
        
        self = object.__new__(cls)
        self.keywords = processed_keywords
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # keywords
        keywords = self.keywords
        repr_parts.append(' keywords=[')
        if (keywords is not None):
            index = 0
            limit = len(keywords)
            while True:
                keyword = keywords[index]
                repr_parts.append(repr(keyword))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadata.from_data)
    def from_data(cls, data):
        keyword_array = data.get('keyword_filter', None)
        if (keyword_array is None) or (not keyword_array):
            keywords = None
        else:
            keywords = tuple(sorted(keyword_array))
        
        self = object.__new__(cls)
        self.keywords = keywords
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.to_data)
    def to_data(self):
        data = {}
        
        data['keyword_filter'] = [*self.iter_keywords()]
        
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.keywords != other.keywords:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__hash__)
    def __hash__(self):
        hash_value = 0
        
        keywords = self.keywords
        if (keywords is not None):
            hash_value ^= len(keywords)
            
            for keyword in keywords:
                hash_value ^= hash(keyword)
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.copy)
    def copy(self):
        new = AutoModerationRuleTriggerMetadata.copy(self)
        
        # keywords
        keywords = self.keywords
        if (keywords is not None):
            keywords = tuple(keyword for keyword in keywords)
        new.keywords = keywords
        
        return new
    
    
    def iter_keywords(self):
        """
        Iterates over the keywords of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        keyword : `str`
        """
        keywords = self.keywords
        if (keywords is not None):
            yield from keywords


class KeywordPresetTriggerMetadata(AutoModerationRuleTriggerMetadata):
    """
    Keyword keyword preset trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    excluded_keywords : `None`, `tuple` of `str`
        Excluded keywords from under the rule.
    keyword_presets : `None`, `tuple` of `AutoModerationKeywordPresetType`
        Substrings which will be searched for in content.
    """
    __slots__ = ('excluded_keywords', 'keyword_presets',)
    
    def __new__(cls, keyword_presets, excluded_keywords=None):
        """
        Creates a new keyword preset trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        keyword_presets : `None`, `int`, ``AutoModerationKeywordPresetType``, \
                `iterable` of (`int`, ``AutoModerationKeywordPresetType``)
            Keyword preset defined by Discord which will be searched for in content.
        
        excluded_keywords : `None`, `str`, `iterable` of `str` = `None`, Optional (Keyword only)
            Excluded keywords from the preset.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        # excluded_keywords
        if excluded_keywords is None:
            processed_excluded_keywords = None
        
        elif isinstance(excluded_keywords, str):
            processed_excluded_keywords = (excluded_keywords, )
        
        else:
            iterator = getattr(type(excluded_keywords), '__iter__', None)
            if iterator is None:
                raise TypeError(
                    f'`excluded_keywords` can be `None`, `str`, `iterable` of `str`, '
                    f'got {excluded_keywords.__class__.__name__}; {excluded_keywords!r}.'
                )
            
            processed_excluded_keywords = None
            
            for excluded_keyword in iterator(excluded_keywords):
                if not isinstance(excluded_keyword, str):
                    raise TypeError(
                        f'`excluded_keywords` can contain `str` elements, got {excluded_keyword.__class__.__name__}; {excluded_keyword!r}; '
                        f'excluded_keywords={excluded_keywords!r}'
                    )
                
                if processed_excluded_keywords is None:
                    processed_excluded_keywords = set()
                
                processed_excluded_keywords.add(excluded_keyword)
            
            if (processed_excluded_keywords is not None):
                processed_excluded_keywords = tuple(sorted(processed_excluded_keywords))
        
        # keyword_presets
        if keyword_presets is None:
            processed_keyword_presets = None
        
        elif isinstance(keyword_presets, int):
            keyword_preset = AutoModerationKeywordPresetType.get(keyword_presets)
            processed_keyword_presets = (keyword_preset, )
        
        elif isinstance(keyword_presets, AutoModerationKeywordPresetType):
            processed_keyword_presets = (keyword_presets, )
        
        else:
            iterator = getattr(type(keyword_presets), '__iter__', None)
            if iterator is None:
                raise TypeError(
                    f'`keyword_presets` can be `None`, `int`, `{AutoModerationKeywordPresetType.__name__}`, '
                    f'`iterable` of (`int`, `{AutoModerationKeywordPresetType.__name__}`'
                    f'got {keyword_presets.__class__.__name__}; {keyword_presets!r}.'
                )
            
            processed_keyword_presets = None
            
            for keyword_preset in iterator(keyword_presets):
                if isinstance(keyword_preset, int):
                    keyword_preset = AutoModerationKeywordPresetType.get(keyword_preset)
                
                elif isinstance(keyword_preset, AutoModerationKeywordPresetType):
                    pass
                
                else:
                    raise TypeError(
                        f'`keyword_preset` can contain `int`, `{AutoModerationKeywordPresetType.__name__}` elements, '
                        f'got {keyword_preset.__class__.__name__}; {keyword_preset!r}; '
                        f'keyword_presets={keyword_presets!r}'
                    )
                
                if processed_keyword_presets is None:
                    processed_keyword_presets = set()
                
                processed_keyword_presets.add(keyword_preset)
            
            if (processed_keyword_presets is not None):
                processed_keyword_presets = tuple(sorted(processed_keyword_presets))
        
        
        
        
        self = object.__new__(cls)
        self.excluded_keywords = processed_excluded_keywords
        self.keyword_presets = processed_keyword_presets
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # keyword_presets
        keyword_presets = self.keyword_presets
        repr_parts.append(' keywords=[')
        if (keyword_presets is not None):
            index = 0
            limit = len(keyword_presets)
            while True:
                keyword = keyword_presets[index]
                repr_parts.append(repr(keyword.name))
                repr_parts.append('~')
                repr_parts.append(repr(keyword.value))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        repr_parts.append(' excluded_keywords=[')
        if (excluded_keywords is not None):
            index = 0
            limit = len(excluded_keywords)
            while True:
                excluded_keyword = excluded_keywords[index]
                repr_parts.append(repr(excluded_keyword))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadata.from_data)
    def from_data(cls, data):
        # excluded_keywords
        excluded_keyword_array = data.get('allow_list', None)
        if (excluded_keyword_array is None) or (not excluded_keyword_array):
            excluded_keywords = None
        else:
            excluded_keywords = tuple(sorted(excluded_keyword_array))
        
        # keyword_preset_array
        keyword_preset_array = data.get('presets', None)
        if (keyword_preset_array is None) or (not keyword_preset_array):
            keyword_presets = None
        else:
            keyword_presets = tuple(sorted(
                AutoModerationKeywordPresetType.get(keyword_preset) for keyword_preset in keyword_preset_array
            ))
        
        self = object.__new__(cls)
        self.excluded_keywords = excluded_keywords
        self.keyword_presets = keyword_presets
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.to_data)
    def to_data(self):
        data = {}
        
        # excluded_keywords
        data['allow_list'] = [*self.iter_excluded_keywords()]
        
        # keyword_presets
        data['presets'] = [*self.iter_keyword_presets()]
        
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # excluded_keywords
        if self.excluded_keywords != other.excluded_keywords:
            return False
        
        # keyword_presets
        if self.keyword_presets != other.keyword_presets:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            hash_value ^= len(excluded_keywords)
            
            for excluded_keyword in excluded_keywords:
                hash_value ^= hash(excluded_keyword)
        
        # keyword_presets
        keyword_presets = self.keyword_presets
        if (keyword_presets is not None):
            hash_value ^= len(keyword_presets)
            
            shift = 0
            
            for keyword_preset in keyword_presets:
                shift += 4
                
                hash_value ^= keyword_preset.value << shift
        
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.copy)
    def copy(self):
        new = AutoModerationRuleTriggerMetadata.copy(self)
        
        # keyword_presets
        keyword_presets = self.keyword_presets
        if (keyword_presets is not None):
            keyword_presets = tuple(keyword_preset for keyword_preset in keyword_presets)
        new.keyword_presets = keyword_presets
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            excluded_keywords = tuple(excluded_keyword for excluded_keyword in excluded_keywords)
        new.excluded_keywords = excluded_keywords
        
        return new
    
    
    def iter_excluded_keywords(self):
        """
        Iterates over the excluded keyword of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_keyword : `str`
        """
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            yield from excluded_keywords
    
    
    def iter_keyword_presets(self):
        """
        Iterates over the keyword presets of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        keyword_preset : `AutoModerationKeywordPresetType`
        """
        keyword_presets = self.keyword_presets
        if (keyword_presets is not None):
            yield from keyword_presets



class MentionSpamTriggerMetadata(AutoModerationRuleTriggerMetadata):
    """
    Mention spam trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    mention_limit : `int`
        The amount of mentions in a message after the rule is triggered.
    """
    __slots__ = ('mention_limit',)
    
    def __new__(cls, mention_limit):
        """
        Creates a new mention spam trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        mention_limit : `None`, `int`
            The amount of mentions in a message after the rule is triggered.
            
            Defaults to the allowed maximal amount if given as `None`.
            
        Raises
        ------
        TypeError
            - If `mention_limit` type is incorrect.
        """
        if (mention_limit is None):
            mention_limit = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
        
        elif isinstance(mention_limit, int):
            if mention_limit < 0:
                mention_limit = 0
            
            elif mention_limit > AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX:
                mention_limit = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
            
        else:
            raise TypeError(
                f'`mention_limit` can be `None`, `int`, got {mention_limit.__class__.__name__}; {mention_limit!r}.'
            )
        
        self = object.__new__(cls)
        self.mention_limit = mention_limit
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # mention_limit
        repr_parts.append(' mention_limit=')
        repr_parts.append(repr(self.mention_limit))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadata.from_data)
    def from_data(cls, data):
        mention_limit = data.get('mention_total_limit', None)
        if (mention_limit is None):
            mention_limit = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
        
        self = object.__new__(cls)
        self.mention_limit = mention_limit
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.to_data)
    def to_data(self):
        data = {}
        
        data['mention_total_limit'] = self.mention_limit
        
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.mention_limit != other.mention_limit:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # mention_limit
        hash_value ^= self.mention_limit
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadata.copy)
    def copy(self):
        new = AutoModerationRuleTriggerMetadata.copy(self)
        
        # mention_limit
        new.mention_limit = self.mention_limit
        
        return new



__all__ = ('ComponentRow',)

from scarletio import copy_docs, export, include

from .component_base import ComponentBase
from .debug import _debug_component_components
from .preinstanced import ComponentType


create_component = include('create_component')


@export
class ComponentRow(ComponentBase):
    """
    Action row component.
    
    Attributes
    ----------
    components : `None`, `tuple` of ``ComponentBase``s
        Stored components.
    
    Class Attributes
    ----------------
    custom_id : `NoneType` = `None`
        `custom_id` is not applicable for component rows.
    type : ``ComponentType`` = `ComponentType.row`
        The component's type.
    """
    type = ComponentType.row
    
    __slots__ = ('components',)
    
    def __new__(cls, *components):
        """
        Creates a new action component from the given components.
        
        Parameters
        ----------
        *components : ``ComponentBase``s
            Sub components.
        
        Raises
        ------
        AssertionError
            - If `components` contains a non ``ComponentBase``.
        """
        if __debug__:
            _debug_component_components(components)
        
        # components
        if not components:
            components = None
        
        self = object.__new__(cls)
        self.components = components
        return self
    
    
    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        # components
        component_datas = data.get('components', None)
        if (component_datas is None) or (not component_datas):
            components = None
        else:
            components = [create_component(component_data) for component_data in component_datas]
        self.components = components
        
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        # type
        data = {
            'type' : self.type.value
        }
        
        # components
        components = self.components
        if (components is None):
            component_datas = []
        else:
            component_datas = [component.to_data() for component in components]
        data['components'] = component_datas
        
        return data
    
    
    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : type
        
        # type
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        # sub-component fields : components
        
        # components
        repr_parts.append(', components=')
        components = self.components
        if (components is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(components)
            
            while True:
                component = components[index]
                index += 1
                
                repr_parts.append(repr(component))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # components
        components = self.components
        if (components is not None):
            components = tuple(component.copy() for component in self.components)
        
        new.components = components
        
        return new
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            Keyword parameters referencing attributes.
        
        Other Parameters
        ----------------
        components : `iterable` of ``ComponentBase``s, Optional (Keyword only)
            Sub components.
        
        Returns
        -------
        new : ``ComponentRow``
        """
        # components
        try:
            components = kwargs.pop('components')
        except KeyError:
            components = self.components
        else:
            components = tuple(components)
            
            if __debug__:
                _debug_component_components(components)
            
            if not components:
                components = None
        
        new = object.__new__(type(self))
        new.components = components
        return new
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        # components
        if self.components != other.components:
            return False
        
        return True
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        # type
        hash_value = self.type.value
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components) << 12
            for component in components:
                hash_value ^= hash(component)
        
        return hash_value
    
    
    def __iter__(self):
        """
        Iterates over the direct sub-components of the row.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``ComponentBase``
        """
        components = self.components
        if (components is not None):
            yield from components
    
    
    @copy_docs(ComponentBase._iter_components)
    def _iter_components(self):
        yield self
        
        components = self.components
        if (components is not None):
            for component in components:
                yield from component._iter_components()
    
    
    @copy_docs(ComponentBase._replace_direct_sub_components)
    def _replace_direct_sub_components(self, relation):
        components = self.components
        if (components is not None):
            self.components = tuple(relation.get(component, component) for component in components)
    
    
    @copy_docs(ComponentBase._iter_direct_sub_components)
    def _iter_direct_sub_components(self):
        components = self.components
        if (components is not None):
            yield from components



__all__ = ('InteractionFieldBase',)

class InteractionFieldBase:
    """
    Base class for values assigned to ``InteractionEvent.interaction`` field.
    """
    __slots__ = ()
    
    def __new__(cls, data, interaction_event):
        """
        Creates a new interaction field from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received interaction field data.
        interaction_event : ``InteractionEvent``
            The parent interaction event.
        """
        return None
    
    
    def __repr__(self):
        """Returns the interaction field's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __hash__(self):
        """Returns the interaction field hash value."""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two interaction field are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return False



__all__ = ('InteractionResponseContext',)

from scarletio import Task, to_coroutine

from ..core import KOKORO


RESPONSE_FLAG_DEFERRING = 1 << 0
RESPONSE_FLAG_DEFERRED = 1 << 1
RESPONSE_FLAG_RESPONDING = 1 << 2
RESPONSE_FLAG_RESPONDED = 1 << 3
RESPONSE_FLAG_EPHEMERAL = 1 << 4

RESPONSE_FLAG_NONE = 0
RESPONSE_FLAG_ACKNOWLEDGING = RESPONSE_FLAG_DEFERRING | RESPONSE_FLAG_RESPONDING
RESPONSE_FLAG_ACKNOWLEDGED = RESPONSE_FLAG_DEFERRED | RESPONSE_FLAG_RESPONDED
RESPONSE_FLAG_DEFERRING_OR_DEFERRED = RESPONSE_FLAG_DEFERRING | RESPONSE_FLAG_DEFERRED
RESPONSE_FLAG_RESPONDING_OR_RESPONDED = RESPONSE_FLAG_RESPONDING | RESPONSE_FLAG_RESPONDED
RESPONSE_FLAG_ACKNOWLEDGING_OR_ACKNOWLEDGED = RESPONSE_FLAG_ACKNOWLEDGING | RESPONSE_FLAG_ACKNOWLEDGED


class InteractionResponseContext:
    """
    Interaction response context manager for managing the interaction event's response flag.
    
    Attributes
    ----------
    interaction_event : ``InteractionEvent``
        The respective interaction event event.
    is_deferring : `bool`
        Whether the request just deferring the interaction event.
    is_ephemeral : `bool`
        Whether the request is ephemeral.
    """
    __slots__ = ('interaction_event', 'is_deferring', 'is_ephemeral',)
    
    def __new__(cls, interaction_event, is_deferring, is_ephemeral):
        """
        Creates a new ``InteractionResponseContext`` with the given parameters.
        
        Parameters
        ----------
        is_deferring : `bool`
            Whether the request just deferring the interaction event.
        is_ephemeral : `bool`
            Whether the request is ephemeral.
        """
        self = object.__new__(cls)
        self.interaction_event = interaction_event
        self.is_deferring = is_deferring
        self.is_ephemeral = is_ephemeral
        return self
    
    
    def __repr__(self):
        """Returns the interaction response context's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        if self.is_deferring:
            repr_parts.append(' deferring')
        
        if self.is_ephemeral:
            repr_parts.append(' (ephemeral)')
        
        repr_parts.append(' interaction_event=')
        repr_parts.append(repr(self.interaction_event))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @to_coroutine
    def ensure(self, coroutine):
        """
        Ensures the coroutine within the interaction response context
        
        This method is an awaitable generator.
        
        Parameters
        ----------
        coroutine : ``CoroutineType``
        """
        self.interaction_event._async_task = Task(self._async(coroutine), KOKORO)
        yield # skip a ready cycle
    
    
    async def _async(self, coroutine):
        """
        Ensures the coroutine within the interaction response context. Wrapped into a task by ``.ensure``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        coroutine : ``CoroutineType``
        """
        try:
            
            async with self:
                await coroutine
        
        finally:
            self.interaction_event._async_task = None
    
    
    async def __aenter__(self):
        """
        Enters the context manager as deferring or responding if applicable.
        
        This method is a coroutine.
        """
        interaction_event = self.interaction_event
        await interaction_event._wait_for_async_task_completion()
        
        response_flag = interaction_event._response_flag
        
        if self.is_deferring:
            if not (response_flag & RESPONSE_FLAG_ACKNOWLEDGING_OR_ACKNOWLEDGED):
                response_flag |= RESPONSE_FLAG_DEFERRING
        else:
            if (not response_flag & RESPONSE_FLAG_RESPONDING_OR_RESPONDED):
                response_flag |= RESPONSE_FLAG_RESPONDING
        
        interaction_event._response_flag = response_flag
        
        return self
    
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Exits the context manager, marking the interaction event as deferred or responded if no exception occurred.
        
        This method is a coroutine.
        """
        interaction_event = self.interaction_event
        response_flag = interaction_event._response_flag
        if exc_type is None:
            if self.is_ephemeral:
                if not response_flag & RESPONSE_FLAG_ACKNOWLEDGED:
                    response_flag ^= RESPONSE_FLAG_EPHEMERAL
            
            if self.is_deferring:
                if response_flag & RESPONSE_FLAG_DEFERRING:
                    response_flag ^= RESPONSE_FLAG_DEFERRING
                    response_flag |= RESPONSE_FLAG_DEFERRED
            else:
                if response_flag & RESPONSE_FLAG_RESPONDING:
                    response_flag ^= RESPONSE_FLAG_RESPONDING
                    response_flag |= RESPONSE_FLAG_RESPONDED
                    
                    if response_flag & RESPONSE_FLAG_DEFERRED:
                        response_flag ^= RESPONSE_FLAG_DEFERRED

        
        else:
            if self.is_deferring:
                if response_flag & RESPONSE_FLAG_DEFERRING:
                    response_flag ^= RESPONSE_FLAG_DEFERRING
            else:
                if response_flag & RESPONSE_FLAG_RESPONDING:
                    response_flag ^= RESPONSE_FLAG_RESPONDING
        
        interaction_event._response_flag = response_flag
        return False



__all__ = ('ChannelMetadataGuildThreadPublic',)

from scarletio import copy_docs

from ...preconverters import preconvert_flag

from .. import channel_types as CHANNEL_TYPES
from ..flags import ChannelFlag

from .guild_thread_base import ChannelMetadataGuildThreadBase


class ChannelMetadataGuildThreadPublic(ChannelMetadataGuildThreadBase):
    """
    Base guild channel metadata type.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    _created_at : `None`, `datetime`
        When the channel was created.
    archived : `bool`
        Whether the thread s archived.
    archived_at : `None`, `datetime`
        When the thread's archive status was last changed.
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    open : `bool`
        Whether the thread channel is open.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channel` permissions are unaffected.
    thread_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        The users inside of the thread if any.
    type : `int` = `12`
        The channel's Discord side type.
    owner_id : `int`
        The channel's creator's identifier. Defaults to `0`.
    flags : ``ChannelFlag``
        The channel's flags.
    
    Class Attributes
    ----------------
    type : `int` = `CHANNEL_TYPES.guild_thread_public`
        The channel's type.
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('flags',)
    
    type = CHANNEL_TYPES.guild_thread_public


    @copy_docs(ChannelMetadataGuildThreadBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if not ChannelMetadataGuildThreadBase._compare_attributes_to(self, other):
            return False
        
        if self.flags != other.flags:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildThreadBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildThreadPublic, cls)._create_empty()
        
        self.flags = ChannelFlag()
        
        return self


    @copy_docs(ChannelMetadataGuildThreadBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildThreadBase._update_attributes(self, data)
        
        self.flags = ChannelFlag(data.get('flags', 0))
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildThreadBase._difference_update_attributes(self, data)
        
        flags = data.get('flags', 0)
        if (self.flags != flags):
            flags = ChannelFlag(flags)
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        return old_attributes
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildThreadBase._precreate)
    def _precreate(cls, keyword_parameters):
        self = super(ChannelMetadataGuildThreadPublic, cls)._precreate(keyword_parameters)
        
        try:
            flags = keyword_parameters.pop('flags')
        except KeyError:
            pass
        else:
            flags = preconvert_flag(flags, 'flags', ChannelFlag)
            self.flags = flags
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._to_data)
    def _to_data(self):
        data = ChannelMetadataGuildThreadBase._to_data(self)
        
        data['flags'] = self.flags
        
        return data



__all__ = ('InteractionEvent',)

from scarletio import Future, export, future_or_timeout, include, shield

from ...bases import DiscordEntity, EventBase
from ...channel import create_partial_channel_from_id
from ...core import (
    APPLICATION_ID_TO_CLIENT, GUILDS, INTERACTION_EVENT_MESSAGE_WAITERS, INTERACTION_EVENT_RESPONSE_WAITERS, KOKORO
)
from ...localizations.utils import get_locale
from ...message import Message
from ...permission import Permission
from ...permission.permission import PERMISSION_PRIVATE
from ...user import ClientUserBase, User
from ...utils import now_as_id, seconds_to_id_difference

from ..interaction_response_context import (
    RESPONSE_FLAG_ACKNOWLEDGED, RESPONSE_FLAG_ACKNOWLEDGING, RESPONSE_FLAG_DEFERRED, RESPONSE_FLAG_DEFERRING,
    RESPONSE_FLAG_EPHEMERAL, RESPONSE_FLAG_NONE, RESPONSE_FLAG_RESPONDED, RESPONSE_FLAG_RESPONDING
)

from .application_command_autocomplete_interaction import ApplicationCommandAutocompleteInteraction
from .application_command_interaction import ApplicationCommandInteraction
from .component_interaction import ComponentInteraction
from .form_submit_interaction import FormSubmitInteraction
from .preinstanced import InteractionType


create_partial_guild_from_id = include('create_partial_guild_from_id')


INTERACTION_TYPE_TABLE = {
    InteractionType.ping.value: None,
    InteractionType.application_command.value: ApplicationCommandInteraction,
    InteractionType.message_component.value: ComponentInteraction,
    InteractionType.application_command_autocomplete.value: ApplicationCommandAutocompleteInteraction,
    InteractionType.form_submit: FormSubmitInteraction,
}


INTERACTION_EVENT_EXPIRE_AFTER = 900 # 15 min
INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE = seconds_to_id_difference(INTERACTION_EVENT_EXPIRE_AFTER)

@export
class InteractionEvent(DiscordEntity, EventBase, immortal=True):
    """
    Represents a processed `INTERACTION_CREATE` dispatch event.
    
    Attributes
    ----------
    id : `int`
        The interaction's id.
    _async_task : `None`, ``Task``
        Task set if interaction event is acknowledged asynchronously.
    _cached_users : `None`, `list` of ``ClientUserBase``
        A list of users, which are temporary cached.
    _response_flag : `bool`
        The response order state of ``InteractionEvent``
        
        +-------------------------------+-------+---------------------------------------------------+
        | Respective name               | Shift | Description                                       |
        +===============================+=======+===================================================+
        | RESPONSE_FLAG_DEFERRING       | 0     | The vent is being acknowledged.                   |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_DEFERRED        | 1     | The event was acknowledged and response will be   |
        |                               |       | sent later. Shows loading screen for the user.    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDING      | 2     | Responding to the interaction.                    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDED       | 3     | Response was sent on the interaction.             |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_EPHEMERAL       | 4     | Whether the main response is an ephemeral,        |
        |                               |       | showing for the invoking user only.               |
        +-------------------------------+-------+---------------------------------------------------+
        
        Can be used by extensions and is used by the the ``Client``-s to ensure correct flow order.
    application_id : `int`
        The interaction's application's identifier.
    application_permissions : ``Permission``
        The permissions granted to the application in the guild.
    channel_id : `int`
        The channel's identifier from where the interaction was called.
    guild_id : `int`
        The guild's identifier from where the interaction was called from. Might be `0` if the interaction was called
        from a private channel.
    guild_locale : ``Locale``
        The guild's preferred locale if invoked from guild.
    interaction : `None`, ``ApplicationCommandInteraction``, ``ComponentInteraction``, \
            ``ApplicationCommandAutocompleteInteraction``
        
        The called interaction by it's route by the user.
    locale : ``Locale``
        The selected language of the invoking user.
    message : `None`, ``Message``
        The message from where the interaction was received. Applicable for message components.
    token : `str`
        Interaction's token used when responding on it.
    type : ``InteractionType``
        The interaction's type.
    user : ``ClientUserBase``
        The user who called the interaction.
    user_permissions : ``Permission``
        The user's permissions in the respective channel.
    
    Class Attributes
    ----------------
    _USER_GUILD_CACHE : `dict` of (`tuple` (``ClientUserBase``, ``Guild``), `int`)
        A cache which stores `user-guild` pairs as keys and their reference count as values to remember
        ``InteractionEvent``'s ``.user``-s' guild profiles of the respective ``.guild`` even if the ``Guild`` is
        uncached.
    
        Note, that private channel interaction, neither interactions of cached guilds are not added, what means if
        all the clients are kicked from a guild the guild profile can be lost in unexpected time.
    
    Notes
    -----
    The interaction token can be used for 15 minutes, tho if it is not used within the first 3 seconds, it is
    invalidated immediately.
    
    ˙˙InteractionEvent˙˙ instances are weakreferable.
    """
    __slots__ = (
        '_async_task', '_cached_users', '_response_flag', 'application_id', 'application_permissions', 'channel_id',
        'guild_id', 'guild_locale', 'interaction', 'locale', 'message', 'token', 'type', 'user', 'user_permissions'
    )
    
    _USER_GUILD_CACHE = {}
    
    def __new__(cls, data):
        """
        Creates a new ``InteractionEvent`` with the given parameters.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            `INTERACTION_CREATE` dispatch event data.
        """
        # id_
        id_ = int(data['id'])
        
        # application_id
        application_id = int(data['application_id'])
        
        # application_permissions
        application_permissions = data.get('app_permissions', None)
        if (application_permissions is None):
            application_permissions = Permission()
        else:
            application_permissions = Permission(application_permissions)
        
        # channel_id
        channel_id = int(data['channel_id'])
        
        # guild_id
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        if guild_id:
            guild = create_partial_guild_from_id(guild_id)
        else:
            guild = None
        
        # guild_locale
        guild_locale = get_locale(data.get('guild_locale', None))
        
        # interaction
        # We set interaction at the end when the object is fully initialized
        
        # locale
        locale = get_locale(data.get('locale', None))
        
        # message
        try:
            message_data = data['message']
        except KeyError:
            message = None
        else:
            message = Message(message_data)
        
        # token
        token = data['token']
        
        # type
        type_value = data['type']
        type_ = InteractionType.get(type_value)
        
        # user
        try:
            guild_profile_data = data['member']
        except KeyError:
            user_data = data['user']
            guild_profile_data = None
        else:
            user_data = guild_profile_data['user']
        
        user = User.from_data(user_data, guild_profile_data, guild_id)
        
        # user_permissions
        if (guild_profile_data is None):
            user_permissions = PERMISSION_PRIVATE
        else:
            try:
                user_permissions = guild_profile_data['permissions']
            except KeyError:
                user_permissions = PERMISSION_PRIVATE
            else:
                user_permissions = Permission(user_permissions)
        
        
        self = object.__new__(cls)
        self._async_task = None
        self._cached_users = None
        self._response_flag = RESPONSE_FLAG_NONE
        self.id = id_
        self.application_id = application_id
        self.application_permissions = application_permissions
        self.type = type_
        self.channel_id = channel_id
        self.guild_id = guild_id
        self.guild_locale = guild_locale
        self.interaction = None
        self.locale = locale
        self.token = token
        self.user = user
        self.user_permissions = user_permissions
        self.message = message
        
        # Cache our user if called from guild. This is required to kill references to the user in the guild.
        if (guild is not None):
            self._add_cached_user(user)
        
        # All field is set -> we can now create our own child
        interaction_type = INTERACTION_TYPE_TABLE.get(type_value, None)
        if (interaction_type is not None):
            self.interaction = interaction_type(data['data'], self)
        
        # Bind cached users to the guild for un-caching on object unallocation.
        cached_users = self._cached_users
        if (cached_users is not None):
            for user in cached_users:
                key = (user, guild)
                USER_GUILD_CACHE = cls._USER_GUILD_CACHE
                try:
                    reference_count = USER_GUILD_CACHE[key]
                except KeyError:
                    reference_count = 1
                else:
                    reference_count += 1
                
                USER_GUILD_CACHE[key] = reference_count
        
        return self
    
    
    async def wait_for_response_message(self, *, timeout=None):
        """
        Waits for response message. Applicable for application command interactions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional (Keyword only)
            The maximal time to wait for message before `TimeoutError` is raised.
        
        Returns
        -------
        message : ``Message``
            The received message.
        
        Raises
        ------
        RuntimeError
            The interaction was acknowledged with `show_for_invoking_user_only=True` (as ephemeral). Response
            `message` cannot be detected.
        TimeoutError
            Message was not received before timeout.
        """
        message = self.message
        if (message is not None):
            return message
        
        if self._response_flag & RESPONSE_FLAG_EPHEMERAL:
            raise RuntimeError(
                f'The interaction was acknowledged with `show_for_invoking_user_only=True` '
                f'(as ephemeral). Response `message` cannot be detected.'
            )
        
        try:
            waiter = INTERACTION_EVENT_MESSAGE_WAITERS[self]
        except KeyError:
            waiter = Future(KOKORO)
            INTERACTION_EVENT_MESSAGE_WAITERS[self] = waiter
        
        waiter = shield(waiter, KOKORO)
        
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        await waiter
        return self.message
    
    
    def __del__(self):
        """
        Unregisters the user-guild pair from the interaction cache.
        """
        cached_users = self._cached_users
        if cached_users is None:
            return
        
        guild = self.guild
        if (guild is None):
            return
        
        for user in cached_users:
            key = (user, guild)
            USER_GUILD_CACHE = self._USER_GUILD_CACHE
            
            # A client meanwhile joined the guild?
            if not guild.partial:
                try:
                    del USER_GUILD_CACHE[key]
                except KeyError:
                    pass
                return
            
            try:
                reference_count = USER_GUILD_CACHE[key]
            except KeyError:
                reference_count = 0
            else:
                if reference_count == 1:
                    del USER_GUILD_CACHE[key]
                    reference_count = 0
                else:
                    reference_count -= 1
            
            if reference_count == 0:
                try:
                    del user.guild_profiles[guild.id]
                except KeyError:
                    pass
    
    
    def __repr__(self):
        """Returns the representation of the event."""
        repr_parts = ['<', self.__class__.__name__]
        
        response_state_names = None
        response_state = self._response_flag
        
        if response_state == RESPONSE_FLAG_NONE:
            pass
        
        if response_state & RESPONSE_FLAG_DEFERRING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferring')
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferred')
        
        if response_state & RESPONSE_FLAG_RESPONDING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responding')
        
        if response_state & RESPONSE_FLAG_RESPONDED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responded')
        
        if response_state & RESPONSE_FLAG_EPHEMERAL:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('ephemeral')
        
        if (response_state_names is not None):
            repr_parts.append(' (')
            index = 0
            limit = len(response_state_names)
            while True:
                response_state_name = response_state_names[index]
                repr_parts.append(response_state_name)
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append('),')
        
        repr_parts.append(' type=')
        interaction_type = self.type
        repr_parts.append(interaction_type.name)
        repr_parts.append(' (')
        repr_parts.append(repr(interaction_type.value))
        repr_parts.append(')')
        
        
        guild_id = self.guild_id
        if guild_id:
            repr_parts.append(', guild_id=')
            repr_parts.append(repr(guild_id))
            
            repr_parts.append(', application_permissions=')
            repr_parts.append(format(self.application_permissions, 'd'))
        
        
        repr_parts.append(', channel_id=')
        repr_parts.append(repr(self.channel_id))
        
        
        message = self.message
        if (message is not None):
            repr_parts.append(', message=')
            repr_parts.append(repr(message))
        
        
        repr_parts.append(', user=')
        repr_parts.append(repr(self.user))
        
        repr_parts.append(', guild_locale=')
        repr_parts.append(repr(self.guild_locale.name))
        
        if guild_id:
            repr_parts.append(', locale=')
            repr_parts.append(repr(self.locale.name))
        
        repr_parts.append(', interaction=')
        repr_parts.append(repr(self.interaction))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def is_unanswered(self):
        """
        Returns whether the event was not acknowledged and not acknowledging either.
        
        Returns
        -------
        is_unanswered : `bool`
        """
        return True if (self._response_flag == RESPONSE_FLAG_NONE) else False
    
    
    def is_acknowledging(self):
        """
        Returns whether the event is being acknowledged.
        
        Returns
        -------
        is_acknowledging : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_ACKNOWLEDGING) else False
    
    
    def is_acknowledged(self):
        """
        Returns whether the event is acknowledged.
        
        Returns
        -------
        is_acknowledged : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_ACKNOWLEDGED) else False
    
    
    def is_deferred(self):
        """
        Returns whether the event is deferred.
        
        Returns
        -------
        is_deferred : `bool`
        """
        response_state = self._response_flag
        if response_state & RESPONSE_FLAG_RESPONDED:
            return False
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            return True
        
        return False
    
    
    def is_responding(self):
        """
        Returns whether the event it being responded.
        
        Returns
        -------
        is_responding : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_RESPONDING) else False
    
    
    def is_responded(self):
        """
        Returns whether was responded.
        
        Returns
        -------
        is_responded : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_RESPONDED) else False
    
    
    def is_expired(self):
        """
        Returns whether the event is already expired.
        
        Returns
        -------
        is_expired : `bool`
        """
        return now_as_id() > (self.id + INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE)
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 3
    
    
    def __iter__(self):
        """
        Unpacks the event.
        
        This method is a generator.
        """
        yield self.type
        yield self.user
        yield self.interaction
    
    
    @property
    def channel(self):
        """
        Returns the interaction's event's channel.
        
        Returns
        -------
        channel : ``Channel``, `None`
        """
        return create_partial_channel_from_id(self.channel_id, -1, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the interaction's guild.
        
        Returns
        -------
        guild : ``Guild``, `None`
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def client(self):
        """
        Returns the interaction's client.
        
        Returns
        -------
        client : `Client`
        
        Raises
        ------
        RuntimeError
            Client could not be identified.
        """
        try:
            return APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            raise RuntimeError(
                f'Client of {self!r} could not be identified.'
            ) from None
    
    
    @property
    def voice_client(self):
        """
        Returns the voice client of the interaction's client in it's guild.
        
        Returns
        -------
        voice_client : `None`, ``VoiceClient``
        """
        try:
            client = APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            voice_client = None
        else:
            guild_id = self.message.guild_id
            if guild_id:
                voice_client = client.voice_clients.get(guild_id, None)
            else:
                voice_client = None
        
        return voice_client
    
    
    def _add_cached_user(self, user):
        """
        Adds a user to the cached ones by the interaction.
        
        This function might be called inside of ``InteractionEvent.__new__`` when initializing it's interaction.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user to add to cache.
        """
        cached_users = self._cached_users
        if cached_users is None:
            self._cached_users = [user]
        else:
            if (user not in cached_users):
                cached_users.append(user)
    
    
    def _add_response_waiter(self):
        """
        Adds the interaction event to response waiters.
        
        Called when the interaction is application command one, to resolve it's ``.message`` attribute when created.
        """
        INTERACTION_EVENT_RESPONSE_WAITERS[self.id] = self
    
    
    async def _wait_for_async_task_completion(self):
        """
        Waits for async task's completion.
        
        This method is a coroutine.
        """
        async_task = self._async_task
        if (async_task is not None) and (async_task is not KOKORO.current_task):
            try:
                await async_task
            finally:
                self._async_task = None
